#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from __future__ import print_function
from .enums import BerType, PduType, Authentication, Command, AssociationResult, SourceDiagnostic, Service, AcseServiceProvider
from .ConfirmedServiceError import ConfirmedServiceError
from .GXDLMSConfirmedServiceError import GXDLMSConfirmedServiceError
from .internal._GXCommon import _GXCommon
from .GXByteBuffer import GXByteBuffer
from .GXDLMSException import GXDLMSException
from .TranslatorTags import TranslatorTags
from .enums.TranslatorOutputType import TranslatorOutputType
from .TranslatorSimpleTags import TranslatorSimpleTags
from .TranslatorGeneralTags import TranslatorGeneralTags
from .ServiceError import ServiceError
from .TranslatorStandardTags import TranslatorStandardTags
from .enums.Security import Security
from .AesGcmParameter import AesGcmParameter
from .GXCiphering import GXCiphering

#
# The services to access the attributes and methods of COSEM objects are
# determined on DLMS/COSEM Application layer.  The services are carried by
# Application Protocol Data Units (APDUs).
# <p />
# In DLMS/COSEM the meter is primarily a server, and the controlling system
# is
# a client.  Also unsolicited (received without a request) messages are
# available.
# pylint: disable=too-many-public-methods
class _GXAPDU:
    #
    # Retrieves the string that indicates the level of authentication, if any.
    #
    @classmethod
    def getAuthenticationString(cls, settings, data, ignoreAcse):
        if settings.authentication != Authentication.NONE or \
            (not ignoreAcse and settings.cipher and settings.cipher.security != Security.NONE):
            #  Add sender ACSE-requirements field component.
            data.setUInt8(BerType.CONTEXT | PduType.SENDER_ACSE_REQUIREMENTS)
            data.setUInt8(2)
            data.setUInt8(BerType.BIT_STRING | BerType.OCTET_STRING)
            data.setUInt8(0x80)
            data.setUInt8(BerType.CONTEXT | PduType.MECHANISM_NAME)
            #  Len
            data.setUInt8(7)
            #  OBJECT IDENTIFIER
            p = [int(0x60), int(0x85), int(0x74), 0x05, 0x08, 0x02, settings.authentication]
            data.set(p)
        #  If authentication is used.
        if settings.authentication != Authentication.NONE:
            #  Add Calling authentication information.
            len_ = 0
            callingAuthenticationValue = None
            if settings.authentication == Authentication.LOW:
                if settings.password:
                    callingAuthenticationValue = settings.password
            else:
                callingAuthenticationValue = settings.ctoSChallenge
            #  0xAC
            data.setUInt8(BerType.CONTEXT | BerType.CONSTRUCTED | PduType.CALLING_AUTHENTICATION_VALUE)
            #  Len
            if callingAuthenticationValue:
                len_ = len(callingAuthenticationValue)
            data.setUInt8((2 + len_))
            #  Add authentication information.
            data.setUInt8(BerType.CONTEXT)
            #  Len.
            data.setUInt8(len_)
            if len_ != 0:
                data.set(callingAuthenticationValue)

    #
    # Code application context name.
    #
    # @param settings
    #            DLMS settings.
    # @param data
    #            Byte buffer where data is saved.
    # @param cipher
    #            Is ciphering settings.
    #
    @classmethod
    def generateApplicationContextName(cls, settings, data, cipher):
        #  ProtocolVersion
        if settings.protocolVersion:
            len_ = len(settings.protocolVersion)
            data.setUInt8(BerType.CONTEXT | PduType.PROTOCOL_VERSION)
            data.setUInt8(2)
            data.setUInt8(8 - len_)
            _GXCommon.setBitString(data, settings.protocolVersion, False)
        #  Application context name tag
        data.setUInt8((BerType.CONTEXT | BerType.CONSTRUCTED | PduType.APPLICATION_CONTEXT_NAME))
        #  Len
        data.setUInt8(0x09)
        data.setUInt8(BerType.OBJECT_IDENTIFIER)
        #  Len
        data.setUInt8(0x07)
        ciphered = cipher and cipher.isCiphered()
        data.setUInt8(0x60)
        data.setUInt8(0x85)
        data.setUInt8(0x74)
        data.setUInt8(0x5)
        data.setUInt8(0x8)
        data.setUInt8(0x1)
        if settings.getUseLogicalNameReferencing():
            if ciphered:
                data.setUInt8(3)
            else:
                data.setUInt8(1)
        else:
            if ciphered:
                data.setUInt8(4)
            else:
                data.setUInt8(2)

        #  Add system title.
        if not settings.isServer and (ciphered or settings.authentication == Authentication.HIGH_GMAC) or settings.authentication == Authentication.HIGH_ECDSA:
            if len(cipher.systemTitle) != 8:
                raise ValueError("SystemTitle")
            #  Add calling-AP-title
            data.setUInt8((BerType.CONTEXT | BerType.CONSTRUCTED | PduType.CALLING_AP_TITLE))
            #  LEN
            data.setUInt8(2 + len(cipher.systemTitle))
            data.setUInt8(BerType.OCTET_STRING)
            #  LEN
            data.setUInt8(len(cipher.systemTitle))
            data.set(cipher.systemTitle)
        #  Add CallingAEInvocationId.
        if not settings.isServer and settings.userId != -1:
            data.setUInt8(BerType.CONTEXT | BerType.CONSTRUCTED | PduType.CALLING_AE_INVOCATION_ID)
            #  LEN
            data.setUInt8(3)
            data.setUInt8(BerType.INTEGER)
            #  LEN
            data.setUInt8(1)
            data.setUInt8(settings.userId)

    # Reserved for internal use.
    @classmethod
    def getConformanceToArray(cls, data):
        ret = _GXCommon.swapBits(data.getUInt8())
        ret |= _GXCommon.swapBits(data.getUInt8()) << 8
        ret |= _GXCommon.swapBits(data.getUInt8()) << 16
        return ret

    # Reserved for internal use.
    @classmethod
    def setConformanceToArray(cls, value, data):
        data.setUInt8(_GXCommon.swapBits(int(value) & 0xFF))
        data.setUInt8(_GXCommon.swapBits((int(value) >> 8) & 0xFF))
        data.setUInt8(_GXCommon.swapBits((int(value) >> 16) & 0xFF))

    #
    # Generate User information initiate request.
    #
    # @param settings
    #            DLMS settings.
    # @param cipher
    # @param data
    #
    @classmethod
    def getInitiateRequest(cls, settings, data):
        #  Tag for xDLMS-Initiate request
        data.setUInt8(Command.INITIATE_REQUEST)
        #  Usage field for the response allowed component.
        #  Usage field for dedicated-key component.
        if not settings.cipher or settings.cipher.security == Security.NONE or not settings.cipher.dedicatedKey:
            #  Not used
            data.setUInt8(0x00)
        else:
            data.setUInt8(1)
            _GXCommon.setObjectCount(len(settings.cipher.dedicatedKey), data)
            data.set(settings.cipher.dedicatedKey)
        #  encoding of the response-allowed component (BOOLEAN DEFAULT TRUE)
        #  usage flag (FALSE, default value TRUE conveyed)
        data.setUInt8(0)
        #  Usage field of the proposed-quality-of-service component.  Not used
        if settings.qualityOfService == 0:
            data.setUInt8(0x00)
        else:
            data.setUInt8(0x01)
            data.setUInt8(settings.qualityOfService)

        data.setUInt8(settings.dlmsVersion)
        #  Tag for conformance block
        data.setUInt8(0x5F)
        data.setUInt8(0x1F)
        #  length of the conformance block
        data.setUInt8(0x04)
        #  encoding the number of unused bits in the bit string
        data.setUInt8(0x00)
        cls.setConformanceToArray(settings.proposedConformance, data)
        data.setUInt16(settings.maxPduSize)

    #
    # Generate user information.
    #
    # @param settings
    #            DLMS settings.
    # @param cipher
    # @param data
    #            Generated user information.
    #
    @classmethod
    def generateUserInformation(cls, settings, cipher, encryptedData, data):
        data.setUInt8(BerType.CONTEXT | BerType.CONSTRUCTED | PduType.USER_INFORMATION)
        if not cipher or not cipher.isCiphered():
            #  Length for AARQ user field
            data.setUInt8(0x10)
            #  Coding the choice for user-information (Octet STRING, universal)
            data.setUInt8(BerType.OCTET_STRING)
            #  Length
            data.setUInt8(0)
            offset = len(data)
            cls.getInitiateRequest(settings, data)
            data.setUInt8(len(data) - offset, offset - 1)
        else:
            if encryptedData:
                #  Length for AARQ user field
                data.setUInt8(int((4 + len(encryptedData))))
                #  Tag
                data.setUInt8(BerType.OCTET_STRING)
                data.setUInt8(int((2 + len(encryptedData))))
                #  Coding the choice for user-information (Octet STRING,
                #  universal)
                data.setUInt8(int(Command.GLO_INITIATE_REQUEST))
                data.setUInt8(len(encryptedData))
                data.set(encryptedData)
            else:
                tmp = GXByteBuffer()
                cls.getInitiateRequest(settings, tmp)
                p = AesGcmParameter(Command.GLO_INITIATE_REQUEST, cipher.systemTitle, cipher.blockCipherKey, cipher.authenticationKey)
                p.security = cipher.security
                p.invocationCounter = cipher.invocationCounter
                crypted = GXCiphering.encrypt(p, tmp.array())
                cipher.invocationCounter = 1 + cipher.invocationCounter
                #  Length for AARQ user field
                data.setUInt8(2 + len(crypted))
                #  Coding the choice for user-information (Octet string,
                #  universal)
                data.setUInt8(BerType.OCTET_STRING)
                data.setUInt8(len(crypted))
                data.set(crypted)

    #
    # Generates Aarq.
    #
    @classmethod
    def generateAarq(cls, settings, cipher, encryptedData, data):
        #  AARQ APDU Tag
        data.setUInt8(BerType.APPLICATION | BerType.CONSTRUCTED)
        #  Length is updated later.
        offset = len(data)
        data.setUInt8(0)
        # /////////////////////////////////////////
        #  Add Application context name.
        cls.generateApplicationContextName(settings, data, cipher)
        cls.getAuthenticationString(settings, data, encryptedData)
        cls.generateUserInformation(settings, cipher, encryptedData, data)
        data.setUInt8((len(data) - offset - 1), offset)

    # pylint: disable=unused-variable
    @classmethod
    def getConformance(cls, value, xml):
        tmp = 1
        if xml.outputType == TranslatorOutputType.SIMPLE_XML:
            for it in range(0, 24):
                if (tmp & value) != 0:
                    xml.appendLine(TranslatorGeneralTags.CONFORMANCE_BIT, "Name", TranslatorSimpleTags.conformancetoString(tmp))
                tmp = tmp << 1
        else:
            for it in range(0, 24):
                if (tmp & value) != 0:
                    xml.append(TranslatorStandardTags.conformancetoString(tmp) + " ")
                tmp = tmp << 1

    #
    # Parse User Information from PDU.
    #
    @classmethod
    def parseUserInformation(cls, settings, cipher, data, xml):
        len_ = data.getUInt8()
        if len(data) - data.position < len_:
            if xml is None:
                raise ValueError("Not enough data.")
            xml.appendComment("Error: Invalid data size.")
        #  Encoding the choice for user information
        tag = data.getUInt8()
        if tag != 0x4:
            raise ValueError("Invalid tag.")
        len_ = data.getUInt8()
        if len(data) - data.position < len_:
            if not xml:
                raise ValueError("Not enough data.")
            xml.appendComment("Error: Invalid data size.")
        if xml and xml.outputType == TranslatorOutputType.STANDARD_XML:
            xml.appendLine(TranslatorGeneralTags.USER_INFORMATION, None, data.toHex(False, data.position, len_))
            data.position = data.position + len_
            return
        _GXAPDU.parseInitiate(False, settings, cipher, data, xml)

    # pylint: disable=too-many-arguments,too-many-locals,unused-argument
    @classmethod
    def parse(cls, initiateRequest, settings, cipher, data, xml, tag2):
        len_ = 0
        tmp2 = GXByteBuffer()
        tmp2.setUInt8(0)
        response = tag2 == Command.INITIATE_RESPONSE
        if response:
            if xml:
                #  <InitiateResponse>
                xml.appendStartTag(Command.INITIATE_RESPONSE)
            #  Optional usage field of the negotiated quality of service
            #  component
            tag = data.getUInt8()
            if tag != 0:
                len_ = data.getUInt8()
                data.position = data.position + len_
                if len_ == 0 and xml and xml.outputType == TranslatorOutputType.SIMPLE_XML:
                    #  NegotiatedQualityOfService
                    xml.appendLine(TranslatorGeneralTags.NEGOTIATED_QUALITY_OF_SERVICE, "Value", "00")
        elif tag2 == Command.INITIATE_REQUEST:
            if xml:
                xml.appendStartTag(Command.INITIATE_REQUEST)
            #  Optional usage field of the negotiated quality of service
            #  component
            tag = data.getUInt8()
            if tag != 0:
                len_ = data.getUInt8()
                tmp = bytearray(len_)
                data.get(tmp)
                if settings.cipher:
                    settings.cipher.setDedicatedKey(tmp)
                if xml:
                    xml.appendLine(TranslatorGeneralTags.DEDICATED_KEY, None, GXByteBuffer.hex(tmp, False))
            elif settings.cipher:
                settings.cipher.dedicatedKey = None
            #  Optional usage field of the negotiated quality of service
            #  component
            tag = data.getUInt8()
            if tag != 0:
                len_ = data.getUInt8()
                if xml and (initiateRequest or xml.outputType == TranslatorOutputType.SIMPLE_XML):
                    xml.appendLine(TranslatorGeneralTags.PROPOSED_QUALITY_OF_SERVICE, None, str(len_))
            else:
                if xml and xml.outputType == TranslatorOutputType.STANDARD_XML:
                    xml.appendLine(TranslatorTags.RESPONSE_ALLOWED, None, "true")
            #  Optional usage field of the proposed quality of service
            #  component
            tag = data.getUInt8()
            #  Skip if used.
            if tag != 0:
                len_ = data.getUInt8()
                data.position = data.position + len_
        elif tag2 == Command.CONFIRMED_SERVICE_ERROR:
            if xml:
                xml.appendStartTag(Command.CONFIRMED_SERVICE_ERROR)
                if xml.outputType == TranslatorOutputType.STANDARD_XML:
                    data.getUInt8()
                    xml.appendStartTag(TranslatorTags.INITIATE_ERROR)
                    type_ = ServiceError(data.getUInt8())
                    str_ = TranslatorStandardTags.serviceErrorToString(type_)
                    value = TranslatorStandardTags.getServiceErrorValue(type_, int(data.getUInt8()))
                    xml.appendLine("x:" + str_, None, value)
                    xml.appendEndTag(TranslatorTags.INITIATE_ERROR)
                else:
                    xml.appendLine(TranslatorTags.SERVICE, "Value", xml.integerToHex(data.getUInt8(), 2))
                    type_ = ServiceError(data.getUInt8())
                    xml.appendStartTag(TranslatorTags.SERVICE_ERROR)
                    xml.appendLine(TranslatorSimpleTags.serviceErrorToString(type_), "Value", TranslatorSimpleTags.getServiceErrorValue(type_, int(data.getUInt8())))
                    xml.appendEndTag(TranslatorTags.SERVICE_ERROR)
                xml.appendEndTag(Command.CONFIRMED_SERVICE_ERROR)
                return
            raise GXDLMSConfirmedServiceError(ConfirmedServiceError(data.getUInt8()), ServiceError(data.getUInt8()), data.getUInt8())
        else:
            if xml:
                xml.appendComment("Error: Failed to descypt data.")
                data.position = len(data)
                return
            raise ValueError("Invalid tag.")
        #  Get DLMS version number.
        if not response:
            settings.dlmsVersion = data.getUInt8()
            if settings.dlmsVersion != 6:
                if not settings.isServer:
                    raise ValueError("Invalid DLMS version number.")
            #  ProposedDlmsVersionNumber
            if xml and (initiateRequest or xml.outputType == TranslatorOutputType.SIMPLE_XML):
                xml.appendLine(TranslatorGeneralTags.PROPOSED_DLMS_VERSION_NUMBER, "Value", xml.integerToHex(settings.dlmsVersion, 2))
        else:
            if data.getUInt8() != 6:
                raise ValueError("Invalid DLMS version number.")
            if xml and (initiateRequest or xml.outputType == TranslatorOutputType.SIMPLE_XML):
                xml.appendLine(TranslatorGeneralTags.NEGOTIATED_DLMS_VERSION_NUMBER, "Value", xml.integerToHex(settings.dlmsVersion, 2))
        #  Tag for conformance block
        tag = data.getUInt8()
        if tag != 0x5F:
            raise ValueError("Invalid tag.")
        #  Old Way...
        if data.getUInt8(data.position) == 0x1F:
            data.getUInt8()
        len_ = data.getUInt8()
        #  The number of unused bits in the bit string.
        tag = data.getUInt8()
        v = cls.getConformanceToArray(data)
        if settings.isServer:
            settings.negotiatedConformance = v & settings.proposedConformance
            if xml:
                xml.appendStartTag(TranslatorGeneralTags.PROPOSED_CONFORMANCE)
                cls.getConformance(v, xml)
        else:
            if xml:
                xml.appendStartTag(TranslatorGeneralTags.NEGOTIATED_CONFORMANCE)
                cls.getConformance(v, xml)
            settings.negotiatedConformance = v
        if not response:
            #  Proposed max PDU size.
            pdu = data.getUInt16()
            settings.maxPduSize = pdu
            if xml:
                #  ProposedConformance closing
                if xml.outputType == TranslatorOutputType.SIMPLE_XML:
                    xml.appendEndTag(TranslatorGeneralTags.PROPOSED_CONFORMANCE)
                elif initiateRequest:
                    xml.append(TranslatorGeneralTags.PROPOSED_CONFORMANCE, False)
                #  ProposedMaxPduSize
                xml.appendLine(TranslatorGeneralTags.PROPOSED_MAX_PDU_SIZE, "Value", xml.integerToHex(pdu, 4))
            #  If client asks too high PDU.
            if pdu > settings.maxServerPDUSize:
                settings.setMaxPduSize = settings.maxServerPDUSize
        else:
            pdu = data.getUInt16()
            if xml is None and pdu < 64:
                raise GXDLMSConfirmedServiceError(ConfirmedServiceError.INITIATE_ERROR, ServiceError.SERVICE, Service.PDU_SIZE)
            #  Max PDU size.
            settings.maxPduSize = pdu
            if xml:
                #  NegotiatedConformance closing
                if xml.outputType == TranslatorOutputType.SIMPLE_XML:
                    xml.appendEndTag(TranslatorGeneralTags.NEGOTIATED_CONFORMANCE)
                elif initiateRequest:
                    xml.append(TranslatorGeneralTags.NEGOTIATED_CONFORMANCE, False)
                #  NegotiatedMaxPduSize
                xml.appendLine(TranslatorGeneralTags.NEGOTIATED_MAX_PDU_SIZE, "Value", xml.integerToHex(settings.maxPduSize, 4))
        if response:
            #  VAA Name
            tag = data.getUInt16()
            if xml:
                if initiateRequest or xml.outputType == TranslatorOutputType.SIMPLE_XML:
                    xml.appendLine(TranslatorGeneralTags.VAA_NAME, "Value", xml.integerToHex(tag, 4))
            if tag == 0x0007:
                if initiateRequest:
                    settings.setUseLogicalNameReferencing(True)
                else:
                    #  If LN
                    if not settings.getUseLogicalNameReferencing() and xml is None:
                        raise ValueError("Invalid VAA.")
            elif tag == 0xFA00:
                #  If SN
                if initiateRequest:
                    settings.setUseLogicalNameReferencing(False)
                else:
                    if settings.getUseLogicalNameReferencing():
                        raise ValueError("Invalid VAA.")
            else:
                #  Unknown VAA.
                raise ValueError("Invalid VAA.")
            if xml:
                #  <InitiateResponse>
                xml.appendEndTag(Command.INITIATE_RESPONSE)
        elif xml:
            xml.appendEndTag(Command.INITIATE_REQUEST)

    #pylint: disable=too-many-function-args, broad-except, too-many-arguments,
    #too-many-locals
    @classmethod
    def parseInitiate(cls, initiateRequest, settings, cipher, data, xml):
        #  Tag for xDLMS-Initate.response
        tag = data.getUInt8()
        originalPos = 0
        if tag in (Command.GLO_INITIATE_RESPONSE, Command.GLO_INITIATE_REQUEST,
                   Command.DED_INITIATE_RESPONSE, Command.DED_INITIATE_REQUEST,
                   Command.GENERAL_GLO_CIPHERING, Command.GENERAL_DED_CIPHERING):
            if xml:
                originalPos = data.position
                if tag in (Command.GENERAL_GLO_CIPHERING, Command.GENERAL_DED_CIPHERING):
                    cnt = _GXCommon.getObjectCount(data)
                    st = bytearray(cnt)
                    data.get(st)
                else:
                    if tag in (Command.GLO_INITIATE_REQUEST, Command.DED_INITIATE_REQUEST):
                        st = settings.cipher.systemTitle
                    else:
                        st = settings.sourceSystemTitle
                cnt = _GXCommon.getObjectCount(data)
                encrypted = bytearray(cnt)
                data.get(encrypted)
                if st and cipher and cipher.blockCipherKey and cipher.authenticationKey and xml.comments:
                    pos = xml.getXmlLength()
                    try:
                        data.position = originalPos - 1
                        p = AesGcmParameter(0, st, settings.cipher.blockCipherKey, settings.cipher.authenticationKey)
                        p.xml = xml
                        tmp = GXCiphering.decrypt(settings.cipher, p, data)
                        data.clear()
                        data.set(tmp)
                        cipher.setSecurity(p.security)
                        tag1 = data.getUInt8()
                        xml.startComment("Decrypted data:")
                        xml.appendLine("Security: " + str(Security(p.security)))
                        xml.appendLine("Invocation Counter: " + str(p.invocationCounter))
                        cls.parse(initiateRequest, settings, cipher, data, xml, tag1)
                        xml.endComment()
                    except Exception:
                        #  It's OK if this fails.
                        xml.setXmlLength(pos)
                xml.appendLine(tag, None, GXByteBuffer.hex(encrypted, False))
                return
            data.position = data.position - 1
            p = AesGcmParameter(0, settings.sourceSystemTitle, settings.cipher.blockCipherKey, settings.cipher.authenticationKey)
            tmp = GXCiphering.decrypt(settings.cipher, p, data)
            data.size = 0
            data.set(tmp)
            cipher.security = p.security
            cipher.securitySuite = p.securitySuite
            tag = data.getUInt8()
        _GXAPDU.parse(initiateRequest, settings, cipher, data, xml, tag)

    #
    # Parse application context name.
    #
    # @param settings
    #            DLMS settings.
    # @param buff
    #            Received data.
    #
    #pylint: disable=too-many-boolean-expressions, too-many-return-statements
    @classmethod
    def parseApplicationContextName(cls, settings, buff, xml):
        #Get length.
        len_ = buff.getUInt8()
        if len(buff) - buff.position < len_:
            raise ValueError("Encoding failed. Not enough data.")
        if buff.getUInt8() != 0x6:
            raise ValueError("Encoding failed. Not an Object ID.")
        if settings.isServer and settings.cipher:
            settings.cipher.setSecurity(Security.NONE)
        #  Object ID length.
        len_ = buff.getUInt8()
        tmp = bytearray(len_)
        buff.get(tmp)
        if tmp[0] != 0x60 or tmp[1] != 0x85 or tmp[2] != 0x74 or tmp[3] != 0x5 or tmp[4] != 0x8 or tmp[5] != 0x1:
            if xml:
                xml.appendLine(TranslatorGeneralTags.APPLICATION_CONTEXT_NAME, "Value", "UNKNOWN")
                return True
            raise Exception("Encoding failed. Invalid Application context name.")
        name = tmp[6]
        if xml:
            if name == 1:
                if xml.outputType == TranslatorOutputType.SIMPLE_XML:
                    xml.appendLine(TranslatorGeneralTags.APPLICATION_CONTEXT_NAME, "Value", "LN")
                else:
                    xml.appendLine(TranslatorGeneralTags.APPLICATION_CONTEXT_NAME, None, "1")
                settings.setUseLogicalNameReferencing(True)
            elif name == 3:
                if xml.outputType == TranslatorOutputType.SIMPLE_XML:
                    xml.appendLine(TranslatorGeneralTags.APPLICATION_CONTEXT_NAME, "Value", "LN_WITH_CIPHERING")
                else:
                    xml.appendLine(TranslatorGeneralTags.APPLICATION_CONTEXT_NAME, None, "3")
                settings.setUseLogicalNameReferencing(True)
            elif name == 2:
                if xml.outputType == TranslatorOutputType.SIMPLE_XML:
                    xml.appendLine(TranslatorGeneralTags.APPLICATION_CONTEXT_NAME, "Value", "SN")
                else:
                    xml.appendLine(TranslatorGeneralTags.APPLICATION_CONTEXT_NAME, None, "2")
                settings.setUseLogicalNameReferencing(False)
            elif name == 4:
                if xml.outputType == TranslatorOutputType.SIMPLE_XML:
                    xml.appendLine(TranslatorGeneralTags.APPLICATION_CONTEXT_NAME, "Value", "SN_WITH_CIPHERING")
                else:
                    xml.appendLine(TranslatorGeneralTags.APPLICATION_CONTEXT_NAME, None, "4")
                settings.setUseLogicalNameReferencing(False)
            else:
                return False
            return True
        if settings.getUseLogicalNameReferencing():
            if name == 1:
                return True
            #  If ciphering is used.
            return name == 3
        if name == 2:
            return True
        #  If ciphering is used.
        return name == 4

    @classmethod
    def validateAare(cls, settings, buff):
        tag = buff.getUInt8()
        if settings.isServer:
            if tag != (BerType.APPLICATION | BerType.CONSTRUCTED | PduType.PROTOCOL_VERSION):
                raise ValueError("Invalid tag.")
        else:
            if tag != (BerType.APPLICATION | BerType.CONSTRUCTED | PduType.APPLICATION_CONTEXT_NAME):
                raise ValueError("Invalid tag.")

    #
    # Parse APDU.
    #
    @classmethod
    def parsePDU(cls, settings, cipher, buff, xml):
        #  Get AARE tag and length
        cls.validateAare(settings, buff)
        len_ = _GXCommon.getObjectCount(buff)
        size = len(buff) - buff.position
        if len_ > size:
            if xml is None:
                raise ValueError("Not enough data.")
            xml.appendComment("Error: Invalid data size.")
        #  Opening tags
        if xml:
            if settings.isServer:
                xml.appendStartTag(Command.AARQ)
            else:
                xml.appendStartTag(Command.AARE)
        ret = _GXAPDU.parsePDU2(settings, cipher, buff, xml)
        #  Closing tags
        if xml:
            if settings.isServer:
                xml.appendEndTag(Command.AARQ)
            else:
                xml.appendEndTag(Command.AARE)
        return ret

    @classmethod
    def parseProtocolVersion(cls, settings, buff, xml):
        #  Get count.
        buff.getUInt8()
        unusedBits = buff.getUInt8()
        value = buff.getUInt8()
        sb = _GXCommon.toBitString(value, 8 - unusedBits)
        settings.protocolVersion = sb
        if xml:
            xml.appendLine(TranslatorTags.PROTOCOL_VERSION, "Value", settings.protocolVersion)

    #
    # Parse APDU.
    #pylint: disable=broad-except
    @classmethod
    def parsePDU2(cls, settings, cipher, buff, xml):
        resultComponent = AssociationResult.ACCEPTED
        resultDiagnosticValue = SourceDiagnostic.NONE
        len_ = 0
        tag = 0
        while buff.position < len(buff):
            tag = buff.getUInt8()
            if tag == BerType.CONTEXT | BerType.CONSTRUCTED | PduType.APPLICATION_CONTEXT_NAME:
                if not cls.parseApplicationContextName(settings, buff, xml):
                    raise GXDLMSException(AssociationResult.PERMANENT_REJECTED, SourceDiagnostic.NOT_SUPPORTED)
            elif tag == BerType.CONTEXT | BerType.CONSTRUCTED | PduType.CALLED_AP_TITLE:
                #  0xA2
                #  Get length.
                if buff.getUInt8() != 3:
                    raise ValueError("Invalid tag.")
                if settings.isServer:
                    #  Choice for result (INTEGER, universal)
                    if buff.getUInt8() != BerType.OCTET_STRING:
                        raise ValueError("Invalid tag.")
                    len_ = buff.getUInt8()
                    tmp = bytearray(len_)
                    buff.get(tmp)
                    try:
                        settings.sourceSystemTitle = tmp
                    except Exception as ex:
                        if xml is None:
                            raise ex
                    if xml:
                        #  RespondingAPTitle
                        if xml.comments:
                            xml.appendComment(_GXCommon.systemTitleToString(settings.standard, settings.sourceSystemTitle))
                        xml.appendLine(TranslatorTags.CALLED_AP_TITLE, "Value", GXByteBuffer.hex(tmp, False))
                else:
                    #  Choice for result (INTEGER, universal)
                    if buff.getUInt8() != BerType.INTEGER:
                        raise ValueError("Invalid tag.")
                    #  Get length.
                    if buff.getUInt8() != 1:
                        raise ValueError("Invalid tag.")
                    resultComponent = AssociationResult(buff.getUInt8())
                    if xml:
                        if resultComponent != AssociationResult.ACCEPTED:
                            xml.appendComment(str(resultComponent))
                        xml.appendLine(TranslatorGeneralTags.ASSOCIATION_RESULT, "Value", xml.integerToHex(resultComponent, 2))
                        xml.appendStartTag(TranslatorGeneralTags.RESULT_SOURCE_DIAGNOSTIC)
            elif tag == BerType.CONTEXT | BerType.CONSTRUCTED | PduType.CALLED_AE_QUALIFIER:
                #  0xA3
                resultDiagnosticValue = _GXAPDU.parseSourceDiagnostic(settings, buff, xml)
            elif tag == BerType.CONTEXT | BerType.CONSTRUCTED | PduType.CALLED_AP_INVOCATION_ID:
                #  0xA4
                _GXAPDU.parseResult(settings, buff, xml)
            elif tag == BerType.CONTEXT | BerType.CONSTRUCTED | PduType.CALLING_AP_TITLE:
                #  0xA6
                len_ = buff.getUInt8()
                tag = buff.getUInt8()
                len_ = buff.getUInt8()
                tmp = bytearray(len_)
                buff.get(tmp)
                try:
                    settings.setSourceSystemTitle(tmp)
                except Exception as ex:
                    if xml is None:
                        raise ex
                _GXAPDU.appendClientSystemTitleToXml(settings, xml)
            elif tag == BerType.CONTEXT | BerType.CONSTRUCTED | PduType.SENDER_ACSE_REQUIREMENTS:
                #  0xAA
                len_ = buff.getUInt8()
                tag = buff.getUInt8()
                len_ = buff.getUInt8()
                tmp = bytearray(len_)
                buff.get(tmp)
                settings.setStoCChallenge(tmp)
                _GXAPDU.appendServerSystemTitleToXml(settings, xml, tag)
            elif tag == BerType.CONTEXT | BerType.CONSTRUCTED | PduType.CALLING_AE_INVOCATION_ID:
                #  0xA9
                len_ = buff.getUInt8()
                tag = buff.getUInt8()
                len_ = buff.getUInt8()
                settings.userId = buff.getUInt8()
                if xml:
                    #  CallingAPTitle
                    xml.appendLine(TranslatorGeneralTags.CALLING_AE_INVOCATION_ID, "Value", xml.integerToHex(settings.userId, 2))
            elif tag == BerType.CONTEXT | BerType.CONSTRUCTED | PduType.CALLED_AE_INVOCATION_ID:
                #  0xA5
                len_ = buff.getUInt8()
                tag = buff.getUInt8()
                len_ = buff.getUInt8()
                settings.userId = buff.getUInt8()
                if xml:
                    #  CallingAPTitle
                    xml.appendLine(TranslatorGeneralTags.CALLED_AE_INVOCATION_ID, "Value", xml.integerToHex(settings.getUserId(), 2))
            elif tag == BerType.CONTEXT | BerType.CONSTRUCTED | 7:
                #  0xA7
                len_ = buff.getUInt8()
                tag = buff.getUInt8()
                len_ = buff.getUInt8()
                settings.userId = buff.getUInt8()
                if xml:
                    #  CallingAPTitle
                    xml.appendLine(TranslatorGeneralTags.RESPONDING_AE_INVOCATION_ID, "Value", xml.integerToHex(settings.userId, 2))
            elif tag == BerType.CONTEXT | BerType.CONSTRUCTED | PduType.CALLING_AP_INVOCATION_ID:
                #  0xA8
                if buff.getUInt8() != 3:
                    raise ValueError("Invalid tag.")
                if buff.getUInt8() != 2:
                    raise ValueError("Invalid length.")
                if buff.getUInt8() != 1:
                    raise ValueError("Invalid tag length.")
                #  Get value.
                len_ = buff.getUInt8()
                if xml:
                    #  CallingApInvocationId
                    xml.appendLine(TranslatorTags.CALLING_AP_INVOCATION_ID, "Value", xml.integerToHex(len_, 2))
            elif tag in (BerType.CONTEXT | PduType.SENDER_ACSE_REQUIREMENTS, BerType.CONTEXT | PduType.CALLING_AP_INVOCATION_ID):
                #  0x88
                #  Get sender ACSE-requirements field component.
                if buff.getUInt8() != 2:
                    raise ValueError("Invalid tag.")
                if buff.getUInt8() != BerType.OBJECT_DESCRIPTOR:
                    raise ValueError("Invalid tag.")
                #  Get only value because client application is
                #  sending system title with LOW authentication.
                buff.getUInt8()
                if xml:
                    xml.appendLine(tag, "Value", "1")
            elif tag in (BerType.CONTEXT | PduType.MECHANISM_NAME, BerType.CONTEXT | PduType.CALLING_AE_INVOCATION_ID):
                #  0x89
                _GXAPDU.updateAuthentication(settings, buff)
                if xml:
                    if xml.outputType == TranslatorOutputType.SIMPLE_XML:
                        str_ = Authentication.toString(settings.authentication)
                        xml.appendLine(tag, "Value", str_)
                    else:
                        xml.appendLine(tag, "Value", str(settings.authentication))
            elif tag == BerType.CONTEXT | BerType.CONSTRUCTED | PduType.CALLING_AUTHENTICATION_VALUE:
                #  0xAC
                _GXAPDU.updatePassword(settings, buff, xml)
            elif tag == BerType.CONTEXT | BerType.CONSTRUCTED | PduType.USER_INFORMATION:
                #  0xBE
                try:
                    _GXAPDU.parseUserInformation(settings, cipher, buff, xml)
                except Exception:
                    if xml is None:
                        # pylint: disable=raise-missing-from
                        #  Check result component.  Some meters are returning invalid
                        #  user-information if connection failed.
                        if xml is None and resultComponent != AssociationResult.ACCEPTED and resultDiagnosticValue != SourceDiagnostic.NONE:
                            raise GXDLMSException(resultComponent, resultDiagnosticValue)
                        if xml is None and resultComponent != AssociationResult.ACCEPTED and resultDiagnosticValue != AcseServiceProvider.NONE:
                            raise GXDLMSException(resultComponent, resultDiagnosticValue)
                        raise GXDLMSException(AssociationResult.PERMANENT_REJECTED, SourceDiagnostic.NO_REASON_GIVEN)
            elif tag == BerType.CONTEXT:
                #  0x80
                cls.parseProtocolVersion(settings, buff, xml)
            else:
                #  Unknown tags.
                print("Unknown tag: " + str(tag) + ".")
                if buff.position < len(buff):
                    len_ = buff.getUInt8()
                    buff.position = buff.position + len_
        #  All meters don't send user-information if connection is failed.
        #  For this reason result component is check again.
        if xml is None and resultComponent != AssociationResult.ACCEPTED and isinstance(resultDiagnosticValue, (SourceDiagnostic)) and resultDiagnosticValue != SourceDiagnostic.NONE:
            raise GXDLMSException(resultComponent, resultDiagnosticValue)
        return resultDiagnosticValue

    @classmethod
    def parseResult(cls, settings, buff, xml):
        if settings.isServer:
            #  Get len.
            if buff.getUInt8() != 3:
                raise ValueError("Invalid tag.")
            #  Choice for result (Universal, Octetstring type)
            if buff.getUInt8() != BerType.INTEGER:
                raise ValueError("Invalid tag.")
            if buff.getUInt8() != 1:
                raise ValueError("Invalid tag length.")
            #  Get value.
            len_ = buff.getUInt8()
            if xml:
                #  RespondingAPTitle
                xml.appendLine(TranslatorTags.CALLED_AP_INVOCATION_ID, "Value", xml.integerToHex(len_, 2))
        else:
            #  Get length.
            if buff.getUInt8() != 0xA:
                raise ValueError("Invalid tag.")
            #  Choice for result (Universal, Octet string type)
            if buff.getUInt8() != BerType.OCTET_STRING:
                raise ValueError("Invalid tag.")
            #  responding-AP-title-field
            #  Get length.
            len_ = buff.getUInt8()
            tmp = bytearray(len_)
            buff.get(tmp)
            settings.setSourceSystemTitle(tmp)
            cls.appendResultToXml(settings, xml)

    @classmethod
    def parseSourceDiagnostic(cls, settings, buff, xml):
        tag = int()
        resultDiagnosticValue = SourceDiagnostic.NONE
        len_ = buff.getUInt8()
        #  ACSE service user tag.
        tag = buff.getUInt8()
        len_ = buff.getUInt8()
        if settings.isServer:
            calledAEQualifier = bytearray(len_)
            buff.get(calledAEQualifier)
            if xml:
                xml.appendLine(TranslatorTags.CALLED_AE_QUALIFIER, "Value", GXByteBuffer.hex(calledAEQualifier, False))
        else:
            #  Result source diagnostic component.
            tag = buff.getUInt8()
            if tag != BerType.INTEGER:
                raise ValueError("Invalid tag.")
            len_ = buff.getUInt8()
            if len_ != 1:
                raise ValueError("Invalid tag.")
            resultDiagnosticValue = SourceDiagnostic(buff.getUInt8())
            if xml:
                if resultDiagnosticValue != SourceDiagnostic.NONE:
                    xml.appendComment(str(resultDiagnosticValue))
                xml.appendLine(TranslatorGeneralTags.ACSE_SERVICE_USER, "Value", xml.integerToHex(resultDiagnosticValue, 2))
                xml.appendEndTag(TranslatorGeneralTags.RESULT_SOURCE_DIAGNOSTIC)
        return resultDiagnosticValue

    @classmethod
    def appendServerSystemTitleToXml(cls, settings, xml, tag):
        if xml:
            #  RespondingAuthentication
            if xml.outputType == TranslatorOutputType.SIMPLE_XML:
                xml.appendLine(tag, None, GXByteBuffer.hex(settings.getStoCChallenge(), False))
            else:
                xml.append(tag, True)
                xml.append(TranslatorGeneralTags.CHAR_STRING, True)
                xml.append(GXByteBuffer.hex(settings.getStoCChallenge(), False))
                xml.append(TranslatorGeneralTags.CHAR_STRING, False)
                xml.append(tag, False)
                xml.append("\n")

    @classmethod
    def appendClientSystemTitleToXml(cls, settings, xml):
        if xml:
            #  CallingAPTitle
            xml.appendLine(TranslatorGeneralTags.CALLING_AP_TITLE, "Value", GXByteBuffer.hex(settings.sourceSystemTitle, False))

    @classmethod
    def appendResultToXml(cls, settings, xml):
        if xml:
            #  RespondingAPTitle
            xml.appendLine(TranslatorGeneralTags.RESPONDING_AP_TITLE, "Value", GXByteBuffer.hex(settings.sourceSystemTitle, False))

    @classmethod
    def updatePassword(cls, settings, buff, xml):
        tmp = []
        len_ = buff.getUInt8()
        #  Get authentication information.
        if buff.getUInt8() != 0x80:
            raise ValueError("Invalid tag.")
        len_ = buff.getUInt8()
        tmp = bytearray(len_)
        buff.get(tmp)
        if settings.authentication == Authentication.LOW:
            settings.password = tmp
        else:
            settings.ctoSChallenge = tmp
        if xml:
            if xml.outputType == TranslatorOutputType.SIMPLE_XML:
                if GXByteBuffer.isAsciiString(tmp):
                    xml.appendComment(tmp.decode("utf-8"))
                xml.appendLine(TranslatorGeneralTags.CALLING_AUTHENTICATION, "Value", GXByteBuffer.hex(tmp, False))
            else:
                xml.appendStartTag(TranslatorGeneralTags.CALLING_AUTHENTICATION)
                xml.appendStartTag(TranslatorGeneralTags.CHAR_STRING)
                if settings.authentication == Authentication.LOW:
                    xml.append(GXByteBuffer.hex(settings.password, False))
                else:
                    xml.append(GXByteBuffer.hex(settings.ctoSChallenge, False))
                xml.appendEndTag(TranslatorGeneralTags.CHAR_STRING)
                xml.appendEndTag(TranslatorGeneralTags.CALLING_AUTHENTICATION)

    @classmethod
    def updateAuthentication(cls, settings, buff):
        ch = buff.getUInt8()
        if buff.getUInt8() != 0x60:
            raise ValueError("Invalid tag.")
        if buff.getUInt8() != 0x85:
            raise ValueError("Invalid tag.")
        if buff.getUInt8() != 0x74:
            raise ValueError("Invalid tag.")
        if buff.getUInt8() != 0x05:
            raise ValueError("Invalid tag.")
        if buff.getUInt8() != 0x08:
            raise ValueError("Invalid tag.")
        if buff.getUInt8() != 0x02:
            raise ValueError("Invalid tag.")
        ch = buff.getUInt8()
        if ch < 0 or ch > 7:
            raise ValueError("Invalid tag.")
        settings.authentication = ch

    #pylint: disable=too-many-function-args
    @classmethod
    def getUserInformation(cls, settings, cipher):
        data = GXByteBuffer()
        #  Tag for xDLMS-Initiate response
        data.setUInt8(Command.INITIATE_RESPONSE)
        #  Usage field for the response allowed component (not used)
        data.setUInt8(0x00)
        #  DLMS Version Number
        data.setUInt8(6)
        data.setUInt8(0x5F)
        data.setUInt8(0x1F)
        #  length of the conformance block
        data.setUInt8(0x04)
        #  encoding the number of unused bits in the bit string
        data.setUInt8(0x00)
        cls.setConformanceToArray(settings.negotiatedConformance, data)
        data.setUInt16(settings.maxPduSize)
        #  VAA Name VAA name (0x0007 for LN referencing and 0xFA00 for SN)
        if settings.getUseLogicalNameReferencing():
            data.setUInt16(0x0007)
        else:
            data.setUInt16(0xFA00)
        if cipher and cipher.isCiphered():
            p = AesGcmParameter(Command.GLO_INITIATE_RESPONSE, cipher.systemTitle, cipher.blockCipherKey, cipher.authenticationKey)
            p.security = cipher.security
            p.invocationCounter = cipher.invocationCounter
            return GXCiphering.encrypt(p, data.array())

        if settings.increaseInvocationCounterForGMacAuthentication:
            cipher.InvocationCounter += 1
        return data.array()

    #
    # Server generates AARE message.
    # pylint: disable=too-many-arguments
    @classmethod
    def generateAARE(cls, settings, data, result, diagnostic, cipher, errorData, encryptedData):
        offset = len(data)
        #  Set AARE tag and length 0x61
        data.setUInt8(BerType.APPLICATION | BerType.CONSTRUCTED | PduType.APPLICATION_CONTEXT_NAME)
        #  Length is updated later.
        data.setUInt8(0)
        cls.generateApplicationContextName(settings, data, cipher)
        #  Result 0xA2
        data.setUInt8(BerType.CONTEXT | BerType.CONSTRUCTED | BerType.INTEGER)
        data.setUInt8(3)
        #  len
        data.setUInt8(BerType.INTEGER)
        #  Tag
        #  Choice for result (INTEGER, universal)
        data.setUInt8(1)
        #  Len
        data.setUInt8(result)
        #  ResultValue
        #  SourceDiagnostic
        data.setUInt8(0xA3)
        data.setUInt8(5)
        #  len
        data.setUInt8(0xA1)
        #  Tag
        data.setUInt8(3)
        #  len
        data.setUInt8(2)
        #  Tag
        #  Choice for result (INTEGER, universal)
        data.setUInt8(1)
        #  Len
        #  diagnostic
        data.setUInt8(diagnostic)
        #  SystemTitle
        if cipher and (settings.authentication == Authentication.HIGH_GMAC or cipher.isCiphered()):
            data.setUInt8(BerType.CONTEXT | BerType.CONSTRUCTED | PduType.CALLED_AP_INVOCATION_ID)
            data.setUInt8((len(cipher.systemTitle)))
            data.setUInt8(BerType.OCTET_STRING)
            data.setUInt8()
            data.set(cipher.systemTitle)
        #  Add CalledAEInvocationId.
        if settings.userId != -1:
            data.setUInt8(BerType.CONTEXT | BerType.CONSTRUCTED | PduType.CALLED_AE_INVOCATION_ID)
            #  LEN
            data.setUInt8(3)
            data.setUInt8(BerType.INTEGER)
            #  LEN
            data.setUInt8(1)
            data.setUInt8(settings.userId)
        if settings.authentication > Authentication.LOW:
            #  Add server ACSE-requirenents field component.
            data.setUInt8(0x88)
            data.setUInt8(0x02)
            #  Len.
            data.setUInt16(0x0780)
            #  Add tag.
            data.setUInt8(0x89)
            data.setUInt8(0x07)
            #  Len
            data.setUInt8(0x60)
            data.setUInt8(0x85)
            data.setUInt8(0x74)
            data.setUInt8(0x05)
            data.setUInt8(0x08)
            data.setUInt8(0x02)
            data.setUInt8(settings.authentication)
            #  Add tag.
            data.setUInt8(0xAA)
            data.setUInt8((len(settings.stoCChallenge)))
            #  Len
            data.setUInt8(BerType.CONTEXT)
            data.setUInt8(len(settings.stoCChallenge))
            data.set(settings.stoCChallenge)
        if result == AssociationResult.ACCEPTED or not cipher or cipher.security == Security.NONE:
            #  Add User Information
            #  Tag 0xBE
            data.setUInt8(BerType.CONTEXT | BerType.CONSTRUCTED | PduType.USER_INFORMATION)
            if encryptedData:
                tmp2 = GXByteBuffer(2 + len(encryptedData))
                tmp2.setUInt8(Command.GLO_INITIATE_RESPONSE)
                _GXCommon.setObjectCount(len(encryptedData), tmp2)
                tmp2.set(encryptedData)
                tmp = tmp2.array()
            else:
                if errorData:
                    tmp = errorData
                else:
                    tmp = cls.getUserInformation(settings, cipher)
            data.setUInt8(2 + (len(tmp)))
            #  Coding the choice for user-information (Octet STRING, universal)
            data.setUInt8(BerType.OCTET_STRING)
            #  Length
            data.setUInt8(len(tmp))
            data.set(tmp)
        data.setUInt8((len(data) - offset - 2), (offset + 1))



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from .GXDLMSObject import GXDLMSObject
from .IGXDLMSBase import IGXDLMSBase
from ..enums import ErrorCode
from ..internal._GXCommon import _GXCommon
from ..GXByteBuffer import GXByteBuffer
from ..enums import ObjectType, DataType
from .enums import GsmStatus, GsmCircuitSwitchStatus, GsmPacketSwitchStatus
from .GXDLMSGSMCellInfo import GXDLMSGSMCellInfo
from .GXAdjacentCell import GXAdjacentCell
from ..internal._GXLocalizer import _GXLocalizer

# pylint: disable=too-many-instance-attributes
class GXDLMSGSMDiagnostic(GXDLMSObject, IGXDLMSBase):
    """
    Online help:
    http://www.gurux.fi/Gurux.DLMS.Objects.GXDLMSGSMDiagnostic
    """

    def __init__(self, ln="0.0.25.6.0.255", sn=0):
        """
        Constructor.

        ln : Logical Name of the object.
        sn : Short Name of the object.
        """
        GXDLMSObject.__init__(self, ObjectType.GSM_DIAGNOSTIC, ln, sn)
        self.version = 1
        self.cellInfo = GXDLMSGSMCellInfo()
        self.adjacentCells = []
        self.operator = ""
        self.status = GsmStatus.NONE
        self.circuitSwitchStatus = GsmCircuitSwitchStatus.INACTIVE
        self.packetSwitchStatus = GsmPacketSwitchStatus.INACTIVE
        self.captureTime = None

    #
    # Returns collection of attributes to read.  If attribute is static and
    # already read or device is returned HW error it is not returned.
    #
    def getAttributeIndexToRead(self, all_):
        attributes = []
        #  LN is static and read only once.
        if all_ or not self.logicalName:
            attributes.append(1)
        #  Operator
        if all_ or self.canRead(2):
            attributes.append(2)
        #  Status
        if all_ or self.canRead(3):
            attributes.append(3)
        #  CircuitSwitchStatus
        if all_ or self.canRead(4):
            attributes.append(4)
        #  PacketSwitchStatus
        if all_ or self.canRead(5):
            attributes.append(5)
        #  CellInfo
        if all_ or self.canRead(6):
            attributes.append(6)
        #  AdjacentCells
        if all_ or self.canRead(7):
            attributes.append(7)
        #  CaptureTime
        if all_ or self.canRead(8):
            attributes.append(8)
        return attributes

    def getValues(self):
        return [self.logicalName,
                self.operator,
                self.status,
                self.circuitSwitchStatus,
                self.packetSwitchStatus,
                self.cellInfo,
                self.adjacentCells,
                self.captureTime]

    #
    # Returns amount of attributes.
    #
    def getAttributeCount(self):
        return 8

    #
    # Returns amount of methods.
    #
    def getMethodCount(self):
        return 0

    def getNames(self):
        return (_GXLocalizer.gettext("Logical name"),\
            _GXLocalizer.gettext("Operator"),\
            _GXLocalizer.gettext("Status"),\
            _GXLocalizer.gettext("Circuit switch status"),\
            _GXLocalizer.gettext("Packet switch status"),\
            _GXLocalizer.gettext("Cell info"),\
            _GXLocalizer.gettext("Adjacent cells"),\
            _GXLocalizer.gettext("Capture time"))

    def getDataType(self, index):
        if index == 1:
            ret = DataType.OCTET_STRING
        elif index == 2:
            ret = DataType.STRING
        elif index == 3:
            ret = DataType.ENUM
        elif index == 4:
            ret = DataType.ENUM
        elif index == 5:
            ret = DataType.ENUM
        elif index == 6:
            ret = DataType.STRUCTURE
        elif index == 7:
            ret = DataType.ARRAY
        elif index == 8:
            ret = DataType.DATETIME
        else:
            raise ValueError("getDataType failed. Invalid attribute index.")
        return ret
    #
    # Returns value of given attribute.
    #
    def getValue(self, settings, e):
        bb = None
        if e.index == 1:
            ret = _GXCommon.logicalNameToBytes(self.logicalName)
        elif e.index == 2:
            #pylint: disable=bad-option-value,redefined-variable-type
            if self.operator is None:
                ret = None
            else:
                ret = self.operator.encode()
        elif e.index == 3:
            if not self.status:
                ret = 0
            else:
                ret = self.status
        elif e.index == 4:
            ret = self.circuitSwitchStatus
        elif e.index == 5:
            ret = self.packetSwitchStatus
        elif e.index == 6:
            bb = GXByteBuffer()
            bb.setUInt8(DataType.STRUCTURE)
            if self.version == 0:
                bb.setUInt8(4)
                _GXCommon.setData(settings, bb, DataType.UINT16, self.cellInfo.cellId)
            else:
                bb.setUInt8(7)
                _GXCommon.setData(settings, bb, DataType.UINT32, self.cellInfo.cellId)
            _GXCommon.setData(settings, bb, DataType.UINT16, self.cellInfo.locationId)
            _GXCommon.setData(settings, bb, DataType.UINT8, self.cellInfo.signalQuality)
            _GXCommon.setData(settings, bb, DataType.UINT8, self.cellInfo.ber)
            if self.version > 0:
                _GXCommon.setData(settings, bb, DataType.UINT16, self.cellInfo.mobileCountryCode)
                _GXCommon.setData(settings, bb, DataType.UINT16, self.cellInfo.mobileNetworkCode)
                _GXCommon.setData(settings, bb, DataType.UINT32, self.cellInfo.channelNumber)
            ret = bb
        elif e.index == 7:
            bb = GXByteBuffer()
            bb.setUInt8(DataType.ARRAY)
            if self.adjacentCells is None:
                bb.setUInt8(0)
            else:
                bb.setUInt8(len(self.adjacentCells))
            for it in self.adjacentCells:
                bb.setUInt8(DataType.STRUCTURE)
                bb.setUInt8(2)
                if self.version == 0:
                    _GXCommon.setData(settings, bb, DataType.UINT16, it.cellId)
                else:
                    _GXCommon.setData(settings, bb, DataType.UINT32, it.cellId)
                _GXCommon.setData(settings, bb, DataType.UINT8, it.signalQuality)
            ret = bb
        elif e.index == 8:
            ret = self.captureTime
        else:
            e.error = ErrorCode.READ_WRITE_DENIED
        return ret

    #
    # Set value of given attribute.
    #
    def setValue(self, settings, e):
        if e.index == 1:
            self.logicalName = _GXCommon.toLogicalName(e.value)
        elif e.index == 2:
            if isinstance(e.value, bytearray):
                self.operator = str(e.value)
            elif isinstance(self.operator, (str,)):
                self.operator = str(e.value)
            elif self.operator is None:
                self.operator = None
            else:
                e.error = ErrorCode.READ_WRITE_DENIED
        elif e.index == 3:
            self.status = GsmStatus(e.value)
        elif e.index == 4:
            self.circuitSwitchStatus = GsmCircuitSwitchStatus(e.value)
        elif e.index == 5:
            self.packetSwitchStatus = GsmPacketSwitchStatus(e.value)
        elif e.index == 6:
            if e.value:
                self.cellInfo.cellId = e.value[0]
                self.cellInfo.locationId = e.value[1]
                self.cellInfo.signalQuality = e.value[2]
                self.cellInfo.ber = e.value[3]
                if self.version > 0:
                    self.cellInfo.mobileCountryCode = e.value[4]
                    self.cellInfo.mobileNetworkCode = e.value[5]
                    self.cellInfo.channelNumber = e.value[6]
        elif e.index == 7:
            self.adjacentCells = []
            if e.value:
                for it in e.value:
                    ac = GXAdjacentCell()
                    ac.cellId = it[0]
                    ac.signalQuality = it[1]
                    self.adjacentCells.append(ac)
        elif e.index == 8:
            if isinstance(e.value, bytearray):
                self.captureTime = _GXCommon.changeType(settings, e.value, DataType.DATETIME)
            else:
                self.captureTime = e.value
        else:
            e.error = ErrorCode.READ_WRITE_DENIED

    def load(self, reader):
        self.operator = reader.readElementContentAsString("Operator")
        self.status = reader.readElementContentAsInt("Status")
        self.circuitSwitchStatus = reader.readElementContentAsInt("CircuitSwitchStatus")
        self.packetSwitchStatus = reader.readElementContentAsInt("PacketSwitchStatus")
        if reader.isStartElement("CellInfo", True):
            self.cellInfo.cellId = reader.readElementContentAsLong("CellId")
            self.cellInfo.locationId = reader.readElementContentAsInt("LocationId")
            self.cellInfo.signalQuality = reader.readElementContentAsInt("SignalQuality")
            self.cellInfo.ber = reader.readElementContentAsInt("Ber")
            reader.readEndElement("CellInfo")
        self.adjacentCells = []
        if reader.isStartElement("AdjacentCells", True):
            while reader.isStartElement("Item", True):
                it = GXAdjacentCell()
                it.cellId = reader.readElementContentAsLong("CellId")
                it.signalQuality = reader.readElementContentAsInt("SignalQuality")
                self.adjacentCells.append(it)
            reader.readEndElement("AdjacentCells")
        self.captureTime = reader.readElementContentAsDateTime("CaptureTime")

    def save(self, writer):
        writer.writeElementObject("Operator", self.operator)
        writer.writeElementString("Status", int(self.status))
        writer.writeElementString("CircuitSwitchStatus", int(self.circuitSwitchStatus))
        writer.writeElementString("PacketSwitchStatus", int(self.packetSwitchStatus))
        if self.cellInfo:
            writer.writeStartElement("CellInfo")
            writer.writeElementString("CellId", self.cellInfo.cellId)
            writer.writeElementString("LocationId", self.cellInfo.locationId)
            writer.writeElementString("SignalQuality", self.cellInfo.signalQuality)
            writer.writeElementString("Ber", self.cellInfo.ber)
            writer.writeEndElement()
        if self.adjacentCells:
            writer.writeStartElement("AdjacentCells")
            for it in self.adjacentCells:
                writer.writeStartElement("Item")
                writer.writeElementString("CellId", it.cellId)
                writer.writeElementString("SignalQuality", it.signalQuality)
                writer.writeEndElement()
            writer.writeEndElement()
        writer.writeElementString("CaptureTime", self.captureTime)



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http:#www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http:#www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
class GXDLMSRoutingTable:
    """
    G3-PLC 6LoWPAN routing table.
    """

    __destinationAddress = None
    __nextHopAddress = None
    __routeCost = None
    __hopCount = 0
    __weakLinkCount = 0
    __validTime = None

    @property
    def destinationAddress(self):
        """
        Address of the destination.
        """
        return self.__destinationAddress

    @destinationAddress.setter
    def destinationAddress(self, value):
        self.__destinationAddress = value

    @property
    def nextHopAddress(self):
        """
        Address of the next hop on the route towards the destination.
        """
        return self.__nextHopAddress

    @nextHopAddress.setter
    def nextHopAddress(self, value):
        self.__nextHopAddress = value

    @property
    def routeCost(self):
        """
        Cumulative link cost along the route towards the destination.
        """
        return self.__routeCost

    @routeCost.setter
    def routeCost(self, value):
        self.__routeCost = value

    @property
    def hopCount(self):
        """
        Number of hops of the selected route to the destination.
        """
        return self.__hopCount

    @hopCount.setter
    def hopCount(self, value):
        self.__hopCount = value

    @property
    def weakLinkCount(self):
        """
        Number of weak links to destination.
        """
        return self.__weakLinkCount

    @weakLinkCount.setter
    def weakLinkCount(self, value):
        self.__weakLinkCount = value

    @property
    def validTime(self):
        """
        Remaining time in minutes until when this entry in the routing table is considered valid.
        """
        return self.__validTime

    @validTime.setter
    def validTime(self, value):
        self.__validTime = value



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from .GXDLMSObject import GXDLMSObject
from .IGXDLMSBase import IGXDLMSBase
from ..enums import ErrorCode
from ..internal._GXCommon import _GXCommon
from ..GXByteBuffer import GXByteBuffer
from ..enums import ObjectType, DataType
from .enums import ServiceType, MessageType
from .GXDLMSCaptureObject import GXDLMSCaptureObject
from .GXRepetitionDelay import GXRepetitionDelay
from .GXPushConfirmationParameter import GXPushConfirmationParameter
from .GXPushProtectionParameters import GXPushProtectionParameters

from .enums.PushOperationMethod import PushOperationMethod
from .enums.DataProtectionKeyType import DataProtectionKeyType
from .enums.RestrictionType import RestrictionType
from .enums.ProtectionType import ProtectionType
from .enums.DataProtectionIdentifiedKeyType import DataProtectionIdentifiedKeyType
from .enums.DataProtectionWrappedKeyType import DataProtectionWrappedKeyType
from ..internal._GXLocalizer import _GXLocalizer

# pylint: disable=too-many-instance-attributes
class GXDLMSPushSetup(GXDLMSObject, IGXDLMSBase):
    """
    Online help:
    http://www.gurux.fi/Gurux.DLMS.Objects.GXDLMSPushSetup
    """

    def __init__(self, ln="0.7.25.9.0.255", sn=0):
        """
        Constructor.

        ln : Logical Name of the object.
        sn : Short Name of the object.
        """
        # pylint: disable=super-with-arguments
        super(GXDLMSPushSetup, self).__init__(ObjectType.PUSH_SETUP, ln, sn)
        self.version = 2
        self.pushObjectList = []
        self.communicationWindow = []
        self.repetitionDelay2 = GXRepetitionDelay()
        self.confirmationParameters = GXPushConfirmationParameter()
        self.service = ServiceType.TCP
        self.message = MessageType.COSEM_APDU
        self.destination = None
        self.randomisationStartInterval = 0
        self.numberOfRetries = 0
        self.repetitionDelay = 0
        self.portReference = None
        self.pushClientSAP = 0
        self.pushProtectionParameters = []
        self.pushOperationMethod = PushOperationMethod.UNCONFIRMED_FAILURE
        self.confirmationParameters = GXPushConfirmationParameter()
        self.lastConfirmationDateTime = None

    def getValues(self):
        if self.version == 0:
            return [
                self.logicalName,
                self.pushObjectList,
                (self.service, self.destination, self.message),
                self.communicationWindow,
                self.randomisationStartInterval,
                self.numberOfRetries,
                self.repetitionDelay,
            ]
        if self.version == 0:
            return [
                self.logicalName,
                self.pushObjectList,
                (self.service, self.destination, self.message),
                self.communicationWindow,
                self.randomisationStartInterval,
                self.numberOfRetries,
                self.repetitionDelay,
                self.portReference,
                self.pushClientSAP,
                self.pushProtectionParameters,
                self.pushOperationMethod,
                self.confirmationParameters,
                self.lastConfirmationDateTime,
            ]
        return [
            self.logicalName,
            self.pushObjectList,
            (self.service, self.destination, self.message),
            self.communicationWindow,
            self.randomisationStartInterval,
            self.numberOfRetries,
            self.repetitionDelay2,
            self.portReference,
            self.pushClientSAP,
            self.pushProtectionParameters,
            self.pushOperationMethod,
            self.confirmationParameters,
            self.lastConfirmationDateTime,
        ]

    def getPushValues(self, client, values):
        """
        Get received objects from push message.

        values: Received values.
        Returns clone of captured COSEM objects.
        """
        if len(values) != len(self.pushObjectList):
            raise ValueError("Size of the push object list is different than values.")
        pos = 0
        objects = []
        for k, v in self.pushObjectList:
            co = GXDLMSCaptureObject(v.attributeIndex, v.dataIndex)
            objects.append((k, co))
            if v.attributeIndex == 0:
                tmp = values[pos]
                index = 1
                while index <= k.getAttributeCount():
                    client.updateValue(k, index, tmp[index - 1])
                    index = 1 + index
            else:
                client.updateValue(k, v.attributeIndex, values[pos])
            pos = 1 + pos

    def invoke(self, settings, e):
        if e.index != 1:
            e.error = ErrorCode.READ_WRITE_DENIED

    #
    # Activates the push process.
    #
    def activate(self, client):
        return client.method(self, 1, 0, DataType.INT8)

    #
    # Reset the push process.
    #
    def reset(self, client):
        return client.method(self, 2, 0, DataType.INT8)

    def getAttributeIndexToRead(self, all_):
        attributes = []
        #  LN is static and read only once.
        if all_ or not self.logicalName:
            attributes.append(1)
        #  PushObjectList
        if all_ or self.canRead(2):
            attributes.append(2)
        #  SendDestinationAndMethod
        if all_ or self.canRead(3):
            attributes.append(3)
        #  CommunicationWindow
        if all_ or self.canRead(4):
            attributes.append(4)
        #  RandomisationStartInterval
        if all_ or self.canRead(5):
            attributes.append(5)
        #  NumberOfRetries
        if all_ or self.canRead(6):
            attributes.append(6)
        # RepetitionDelay
        if all_ or self.canRead(7):
            attributes.append(7)
        if self.version > 0:
            # PortReference
            if all_ or self.canRead(8):
                attributes.append(8)
            # PushClientSAP
            if all_ or self.canRead(9):
                attributes.append(9)
            # PushProtectionParameters
            if all_ or self.canRead(10):
                attributes.append(10)
            if self.version > 0:
                # PushOperationMethod
                if all_ or self.canRead(11):
                    attributes.append(11)
                # ConfirmationParameters
                if all_ or self.canRead(12):
                    attributes.append(12)
                # LastConfirmationDateTime
                if all_ or self.canRead(13):
                    attributes.append(13)
        return attributes

    #
    # Returns amount of attributes.
    #
    def getAttributeCount(self):
        if self.version == 0:
            return 7
        if self.version == 1:
            return 10
        return 13

    #
    # Returns amount of methods.
    #
    def getMethodCount(self):
        if self.version < 2:
            return 1
        return 2

    def getNames(self):
        return (_GXLocalizer.gettext("Logical name"),\
            _GXLocalizer.gettext("Object list"),\
            _GXLocalizer.gettext("Send destination and method"),\
            _GXLocalizer.gettext("Communication window"),\
            _GXLocalizer.gettext("Randomisation start interval"),\
            _GXLocalizer.gettext("Number of retries"),\
            _GXLocalizer.gettext("Repetition delay"),\
            _GXLocalizer.gettext("Port reference"),\
            _GXLocalizer.gettext("Push client sap"),\
            _GXLocalizer.gettext("Push protection parameters"),\
            _GXLocalizer.gettext("Push operation method"),\
            _GXLocalizer.gettext("Confirmation parameters"),\
            _GXLocalizer.gettext("Last confirmation date time"))

    #
    # Returns method names.
    #
    def getMethodNames(self):
        if self.version < 2:
            return (_GXLocalizer.gettext("Push"),)
        return (_GXLocalizer.gettext("Push"),\
            _GXLocalizer.gettext("Reset"))

    def getDataType(self, index):
        ret = None
        if index == 1:
            ret = DataType.OCTET_STRING
        elif index == 2:
            ret = DataType.ARRAY
        elif index == 3:
            ret = DataType.STRUCTURE
        elif index == 4:
            ret = DataType.ARRAY
        elif index == 5:
            ret = DataType.UINT16
        elif index == 6:
            ret = DataType.UINT8
        elif index == 7:
            if self.version < 2:
                ret = DataType.UINT16
            else:
                ret = DataType.STRUCTURE
        elif self.version > 0:
            # PortReference
            if index == 8:
                ret = DataType.OCTET_STRING
            # PushClientSAP
            elif index == 9:
                ret = DataType.INT8
            # PushProtectionParameters
            elif index == 10:
                return DataType.ARRAY
            elif self.version > 1:
                # PushOperationMethod
                if index == 11:
                    ret = DataType.ENUM
                # ConfirmationParameters
                if index == 12:
                    ret = DataType.STRUCTURE
                # LastConfirmationDateTime
                if index == 13:
                    ret = DataType.DATETIME
        if ret is None:
            raise ValueError("getDataType failed. Invalid attribute index.")
        return ret

    def __getPushObjectList(self, settings):
        buff = GXByteBuffer()
        buff.setUInt8(DataType.ARRAY)
        _GXCommon.setObjectCount(len(self.pushObjectList), buff)
        for k, v in self.pushObjectList:
            buff.setUInt8(DataType.STRUCTURE)
            if self.version < 1:
                buff.setUInt8(4)
                _GXCommon.setData(None, buff, DataType.UINT16, k.objectType)
                _GXCommon.setData(
                    None,
                    buff,
                    DataType.OCTET_STRING,
                    _GXCommon.logicalNameToBytes(k.logicalName),
                )
                _GXCommon.setData(settings, buff, DataType.INT8, v.attributeIndex)
                _GXCommon.setData(settings, buff, DataType.UINT16, v.dataIndex)
            else:
                buff.setUInt8(5 if self.version == 1 else 6)
                _GXCommon.setData(settings, buff, DataType.UINT16, k.objectType)
                _GXCommon.setData(
                    settings,
                    buff,
                    DataType.OCTET_STRING,
                    _GXCommon.logicalNameToBytes(k.logicalName),
                )
                _GXCommon.setData(settings, buff, DataType.INT8, v.attributeIndex)
                _GXCommon.setData(settings, buff, DataType.UINT16, v.dataIndex)
                # restriction_element
                buff.setUInt8(DataType.STRUCTURE)
                buff.setUInt8(2)
                _GXCommon.setData(settings, buff, DataType.ENUM, v.restriction.type)
                if v.restriction.type == RestrictionType.NONE:
                    _GXCommon.setData(settings, buff, DataType.NONE, None)
                elif v.restriction.type == RestrictionType.DATE:
                    buff.setUInt8(DataType.STRUCTURE)
                    buff.setUInt8(2)
                    _GXCommon.setData(
                        settings, buff, DataType.OCTET_STRING, v.restriction.from_
                    )
                    _GXCommon.setData(
                        settings, buff, DataType.OCTET_STRING, v.restriction.to
                    )
                elif v.restriction.type == RestrictionType.ENTRY:
                    buff.setUInt8(DataType.STRUCTURE)
                    buff.setUInt8(2)
                    _GXCommon.setData(
                        settings, buff, DataType.UINT16, v.restriction.from_
                    )
                    _GXCommon.setData(settings, buff, DataType.UINT16, v.restriction.to)
                if self.version > 1:
                    if v.columns:
                        buff.setUInt8(DataType.ARRAY)
                        _GXCommon.setObjectCount(len(v.columns), buff)
                        for k2, v2 in v.columns:
                            buff.setUInt8(DataType.STRUCTURE)
                            buff.setUInt8(4)
                            _GXCommon.setData(
                                settings, buff, DataType.UINT16, k2.objectType
                            )
                            _GXCommon.setData(
                                settings,
                                buff,
                                DataType.OCTET_STRING,
                                _GXCommon.logicalNameToBytes(k2.logicalName),
                            )
                            _GXCommon.setData(
                                settings, buff, DataType.INT8, v2.attributeIndex
                            )
                            _GXCommon.setData(
                                settings, buff, DataType.UINT16, v2.dataIndex
                            )
                    else:
                        buff.setUInt8(DataType.ARRAY)
                        buff.setUInt8(0)

        return buff

    #
    # Returns value of given attribute.
    #
    def getValue(self, settings, e):
        # pylint: disable=bad-option-value,redefined-variable-type
        buff = GXByteBuffer()
        if e.index == 1:
            ret = _GXCommon.logicalNameToBytes(self.logicalName)
        elif e.index == 2:
            ret = self.__getPushObjectList(settings)
        elif e.index == 3:
            buff.setUInt8(DataType.STRUCTURE)
            buff.setUInt8(3)
            _GXCommon.setData(settings, buff, DataType.ENUM, self.service)
            if self.destination:
                _GXCommon.setData(
                    settings, buff, DataType.OCTET_STRING, self.destination.encode()
                )
            else:
                _GXCommon.setData(settings, buff, DataType.OCTET_STRING, None)
            _GXCommon.setData(settings, buff, DataType.ENUM, self.message)
            ret = buff
        elif e.index == 4:
            buff.setUInt8(DataType.ARRAY)
            _GXCommon.setObjectCount(len(self.communicationWindow), buff)
            for k, v in self.communicationWindow:
                buff.setUInt8(DataType.STRUCTURE)
                buff.setUInt8(2)
                _GXCommon.setData(settings, buff, DataType.OCTET_STRING, k)
                _GXCommon.setData(settings, buff, DataType.OCTET_STRING, v)
            return buff
        elif e.index == 5:
            ret = self.randomisationStartInterval
        elif e.index == 6:
            ret = self.numberOfRetries
        elif e.index == 7:
            if self.version < 2:
                ret = self.repetitionDelay
            else:
                buff.setUInt8(DataType.STRUCTURE)
                _GXCommon.setObjectCount(3, buff)
                _GXCommon.setData(
                    None, buff, DataType.UINT16, self.repetitionDelay2.min
                )
                _GXCommon.setData(
                    None, buff, DataType.UINT16, self.repetitionDelay2.exponent
                )
                _GXCommon.setData(
                    None, buff, DataType.UINT16, self.repetitionDelay2.max
                )
                ret = buff.array()
        elif e.index == 8:
            if self.portReference:
                ret = _GXCommon.logicalNameToBytes(self.portReference.LogicalName)
            else:
                ret = None
        elif e.index == 9:
            ret = self.pushClientSAP
        elif e.index == 10:
            buff.setUInt8(DataType.ARRAY)
            _GXCommon.setObjectCount(len(self.pushProtectionParameters), buff)
            for it in self.pushProtectionParameters:
                buff.setUInt8(DataType.STRUCTURE)
                buff.setUInt8(2)
                _GXCommon.setData(None, buff, DataType.ENUM, it.protectionType)
                buff.setUInt8(DataType.STRUCTURE)
                buff.setUInt8(5)
                _GXCommon.setData(
                    settings, buff, DataType.OCTET_STRING, it.transactionId
                )
                _GXCommon.setData(
                    settings, buff, DataType.OCTET_STRING, it.originatorSystemTitle
                )
                _GXCommon.setData(
                    settings, buff, DataType.OCTET_STRING, it.recipientSystemTitle
                )
                _GXCommon.setData(
                    settings, buff, DataType.OCTET_STRING, it.otherInformation
                )
                buff.setUInt8(DataType.STRUCTURE)
                buff.setUInt8(2)
                _GXCommon.setData(
                    settings, buff, DataType.ENUM, it.keyInfo.dataProtectionKeyType
                )
                buff.setUInt8(DataType.STRUCTURE)
                if it.keyInfo.dataProtectionKeyType == DataProtectionKeyType.IDENTIFIED:
                    buff.setUInt8(1)
                    _GXCommon.setData(
                        settings, buff, DataType.ENUM, it.keyInfo.identifiedKey.keyType
                    )
                elif it.keyInfo.dataProtectionKeyType == DataProtectionKeyType.WRAPPED:
                    buff.setUInt8(2)
                    _GXCommon.setData(
                        settings, buff, DataType.ENUM, it.keyInfo.wrappedKey.keyType
                    )
                    _GXCommon.setData(
                        settings, buff, DataType.OCTET_STRING, it.keyInfo.wrappedKey.key
                    )
                elif it.keyInfo.dataProtectionKeyType == DataProtectionKeyType.AGREED:
                    buff.setUInt8(2)
                    _GXCommon.setData(
                        settings,
                        buff,
                        DataType.OCTET_STRING,
                        it.keyInfo.agreedKey.parameters,
                    )
                    _GXCommon.setData(
                        settings, buff, DataType.OCTET_STRING, it.keyInfo.agreedKey.data
                    )
            ret = buff
        elif e.index == 11:
            ret = self.pushOperationMethod
        elif e.index == 12:
            buff.setUInt8(DataType.STRUCTURE)
            _GXCommon.setObjectCount(2, buff)
            _GXCommon.setData(
                settings, buff, DataType.DATETIME, self.confirmationParameters.startDate
            )
            _GXCommon.setData(
                settings, buff, DataType.UINT32, self.confirmationParameters.interval
            )
            ret = buff.array()
        elif e.index == 13:
            ret = self.lastConfirmationDateTime
        else:
            e.error = ErrorCode.READ_WRITE_DENIED
        return ret

    def __setPushObject(self, settings, e):
        # pylint: disable=import-outside-toplevel
        from .._GXObjectFactory import _GXObjectFactory

        self.pushObjectList = []
        if e.value:
            for it in e.value:
                type_ = it[0]
                ln = _GXCommon.toLogicalName(it[1])
                obj = settings.objects.findByLN(type_, ln)
                if not obj:
                    obj = _GXObjectFactory.createObject(type_)
                    obj.logicalName = ln
                co = GXDLMSCaptureObject()
                co.attributeIndex = it[2]
                co.dataIndex = it[3]
                self.pushObjectList.append((obj, co))
            if self.version > 1:
                restriction = it[4]
                co.restriction.type = RestrictionType(restriction[0])
                if co.restriction.type == RestrictionType.NONE:
                    pass
                elif (
                    co.restriction.type == RestrictionType.DATE
                    or co.restriction.type == RestrictionType.ENTRY
                ):
                    co.restriction.from_ = restriction[1]
                    co.restriction.to = restriction[2]
                else:
                    raise ValueError("Invalid restriction type.")
                for c in it[5]:
                    tp = ObjectType(c[0])
                    ln = _GXCommon.toLogicalName(c[1])
                    obj = settings.objects.findByLN(tp, ln)
                    if obj is None:
                        obj = _GXObjectFactory.createObject(tp)
                        obj.logicalName = ln
                    co.columns.append((obj, co))

    def __getPushProtectionParameters(self, e):
        self.pushProtectionParameters.clear()
        if e.value:
            for it in e.value:
                p = GXPushProtectionParameters()
                p.protectionType = ProtectionType(int(it[0]))
                options = it[1]
                p.transactionId = options[0]
                p.originatorSystemTitle = options[1]
                p.recipientSystemTitle = options[2]
                p.otherInformation = options[3]
                keyInfo = options[4]
                p.keyInfo.dataProtectionKeyType = DataProtectionKeyType(int(keyInfo[0]))
                data = keyInfo[1]
                if p.keyInfo.dataProtectionKeyType == DataProtectionKeyType.IDENTIFIED:
                    p.keyInfo.identifiedKey.keyType = DataProtectionIdentifiedKeyType(
                        int(data[0])
                    )
                elif p.keyInfo.dataProtectionKeyType == DataProtectionKeyType.WRAPPED:
                    p.keyInfo.wrappedKey.keyType = DataProtectionWrappedKeyType(
                        int(data[0])
                    )
                    p.keyInfo.wrappedKey.key = data[1]
                elif p.keyInfo.dataProtectionKeyType == DataProtectionKeyType.AGREED:
                    p.keyInfo.agreedKey.parameters = data[0]
                    p.keyInfo.agreedKey.data = data[1]
                self.pushProtectionParameters.append(p)

    def setValue(self, settings, e):
        if e.index == 1:
            self.logicalName = _GXCommon.toLogicalName(e.value)
        elif e.index == 2:
            self.__setPushObject(settings, e)
        elif e.index == 3:
            # pylint: disable=broad-except
            if e.value:
                self.service = ServiceType(e.value[0])
                try:
                    if self.service == ServiceType.HDLC:
                        self.destination = _GXCommon.toLogicalName(e.value[1])
                    else:
                        # If destination is not ASCII string.
                        if e.value[1] and not GXByteBuffer.isAsciiString(e.value[1]):
                            self.destination = GXByteBuffer.toHex(self.destination)
                        else:
                            self.destination = e.value[1].decode()
                except Exception:
                    self.destination = GXByteBuffer.hex(e.value[1])
                self.message = e.value[2]
        elif e.index == 4:
            self.communicationWindow = []
            if e.value:
                for it in e.value:
                    start = _GXCommon.changeType(settings, it[0], DataType.DATETIME)
                    end = _GXCommon.changeType(settings, it[1], DataType.DATETIME)
                    self.communicationWindow.append((start, end))
        elif e.index == 5:
            self.randomisationStartInterval = e.value
        elif e.index == 6:
            self.numberOfRetries = e.value
        elif e.index == 7:
            if self.version < 2:
                self.repetitionDelay = e.value
            else:
                if e.value:
                    self.repetitionDelay2.min = e.value[0]
                    self.repetitionDelay2.exponent = e.value[1]
                    self.repetitionDelay2.max = e.value[2]
        elif self.version > 0 and e.index == 8:
            self.portReference = None
            if e.value:
                ln = _GXCommon.toLogicalName(e.value)
                self.portReference = settings.objects.findByLN(ObjectType.NONE, ln)
        elif self.version > 0 and e.index == 9:
            self.pushClientSAP = e.value
        elif self.version > 0 and e.index == 10:
            self.__getPushProtectionParameters(e)
        elif self.version > 0 and e.index == 11:
            self.pushOperationMethod = PushOperationMethod(e.value)
        elif self.version > 0 and e.index == 12:
            if e.value:
                self.confirmationParameters.startDate = e.value[0]
                self.confirmationParameters.interval = e.value[1]
            else:
                e.error = ErrorCode.READ_WRITE_DENIED
        elif self.version > 0 and e.index == 13:
            self.lastConfirmationDateTime = e.value
        else:
            e.error = ErrorCode.READ_WRITE_DENIED

    def load(self, reader):
        from .._GXObjectFactory import _GXObjectFactory

        self.pushObjectList = []
        if reader.isStartElement("ObjectList", True):
            while reader.isStartElement("Item", True):
                ot = ObjectType(reader.readElementContentAsInt("ObjectType"))
                ln = reader.readElementContentAsString("LN")
                ai = reader.readElementContentAsInt("AI")
                di = reader.readElementContentAsInt("DI")
                reader.readEndElement("ObjectList")
                co = GXDLMSCaptureObject(ai, di)
                obj = reader.objects.findByLN(ot, ln)
                if not obj:
                    obj = _GXObjectFactory.createObject(ot)
                    obj.LogicalName = ln
                self.pushObjectList.append((obj, co))
            reader.readEndElement("ObjectList")
        self.service = ServiceType(reader.readElementContentAsInt("Service"))
        self.destination = reader.readElementContentAsString("Destination")
        self.message = MessageType(reader.readElementContentAsInt("Message"))
        self.communicationWindow = []
        if reader.isStartElement("CommunicationWindow", True):
            while reader.isStartElement("Item", True):
                start = reader.readElementContentAsDateTime("Start")
                end = reader.readElementContentAsDateTime("End")
                self.communicationWindow.append((start, end))
            reader.readEndElement("CommunicationWindow")
        self.randomisationStartInterval = reader.readElementContentAsInt(
            "RandomisationStartInterval"
        )
        self.numberOfRetries = reader.readElementContentAsInt("NumberOfRetries")
        if self.version < 2:
            self.repetitionDelay = reader.readElementContentAsInt("RepetitionDelay")
        else:
            if reader.isStartElement("RepetitionDelay", True):
                self.repetitionDelay2.min = reader.readElementContentAsInt("Min")
                self.repetitionDelay2.exponent = reader.readElementContentAsInt(
                    "Exponent"
                )
                self.repetitionDelay2.max = reader.readElementContentAsInt("Max")
            reader.readEndElement("RepetitionDelay")
        if self.version > 0:
            ln = reader.readElementContentAsString("PortReference")
            self.portReference = reader.objects.findByLN(ObjectType.NONE, ln)
            if self.portReference is None:
                self.portReference = _GXObjectFactory.createObject(
                    ObjectType.IEC_HDLC_SETUP
                )
                self.portReference.logicalName = ln
            self.pushClientSAP = reader.readElementContentAsInt("PushClientSAP")
            if reader.isStartElement("PushProtectionParameters", True):
                self.pushProtectionParameters.clear()
                while reader.isStartElement("Item", True):
                    it = GXPushProtectionParameters()
                    it.protectionType = ProtectionType(
                        reader.readElementContentAsInt("ProtectionType")
                    )
                    it.transactionId = GXByteBuffer.hexToBytes(
                        reader.readElementContentAsString("TransactionId")
                    )
                    it.originatorSystemTitle = GXByteBuffer.hexToBytes(
                        reader.readElementContentAsString("OriginatorSystemTitle")
                    )
                    it.recipientSystemTitle = GXByteBuffer.hexToBytes(
                        reader.readElementContentAsString("RecipientSystemTitle")
                    )
                    it.otherInformation = GXByteBuffer.hexToBytes(
                        reader.readElementContentAsString("OtherInformation")
                    )
                    it.keyInfo.dataProtectionKeyType = DataProtectionKeyType(
                        reader.readElementContentAsInt("DataProtectionKeyType")
                    )
                    it.keyInfo.identifiedKey.keyType = DataProtectionIdentifiedKeyType(
                        reader.readElementContentAsInt("IdentifiedKey")
                    )
                    it.keyInfo.wrappedKey.keyType = DataProtectionWrappedKeyType(
                        reader.readElementContentAsInt("WrappedKeyType")
                    )
                    it.keyInfo.wrappedKey.key = GXByteBuffer.hexToBytes(
                        reader.readElementContentAsString("WrappedKey")
                    )
                    it.keyInfo.agreedKey.parameters = GXByteBuffer.hexToBytes(
                        reader.readElementContentAsString("WrappedKeyParameters")
                    )
                    it.keyInfo.agreedKey.data = GXByteBuffer.hexToBytes(
                        reader.readElementContentAsString("AgreedKeyData")
                    )
                    self.pushProtectionParameters.append(it)
                reader.readEndElement("PushProtectionParameters")
            if self.version > 1:
                self.pushOperationMethod = PushOperationMethod(
                    reader.readElementContentAsInt("PushOperationMethod")
                )
                self.confirmationParameters.startDate = (
                    reader.readElementContentAsDateTime(
                        "ConfirmationParametersStartDate"
                    )
                )
                self.confirmationParameters.interval = reader.readElementContentAsInt(
                    "ConfirmationParametersInterval"
                )
                self.lastConfirmationDateTime = reader.readElementContentAsDateTime(
                    "LastConfirmationDateTime"
                )

    def save(self, writer):
        if self.pushObjectList:
            writer.writeStartElement("ObjectList")
            for k, v in self.pushObjectList:
                writer.writeStartElement("Item")
                writer.writeElementString("ObjectType", int(k.objectType))
                writer.writeElementString("LN", k.logicalName)
                writer.writeElementString("AI", v.attributeIndex)
                writer.writeElementString("DI", v.dataIndex)
                writer.writeEndElement()
            writer.writeEndElement()
        writer.writeElementString("Service", int(self.service))
        writer.writeElementString("Destination", self.destination)
        writer.writeElementString("Message", int(self.message))
        writer.writeStartElement("CommunicationWindow")
        if self.communicationWindow:
            for k, v in self.communicationWindow:
                writer.writeStartElement("Item")
                writer.writeElementString("Start", k)
                writer.writeElementString("End", v)
                writer.writeEndElement()
        writer.writeEndElement()
        writer.writeElementString(
            "RandomisationStartInterval", self.randomisationStartInterval
        )
        writer.writeElementString("NumberOfRetries", self.numberOfRetries)
        if self.version < 2:
            writer.writeElementString("RepetitionDelay", self.repetitionDelay)
        else:
            writer.writeStartElement("RepetitionDelay")
            writer.writeElementString("Min", self.repetitionDelay2.min)
            writer.writeElementString("Exponent", self.repetitionDelay2.exponent)
            writer.writeElementString("Max", self.repetitionDelay2.max)
            writer.writeEndElement()
        if self.version > 0:
            if self.portReference:
                writer.writeElementString(
                    "PortReference", self.portReference.logicalName
                )
            writer.writeElementString("PushClientSAP", self.pushClientSAP)
            if self.pushProtectionParameters:
                writer.writeStartElement("PushProtectionParameters")
                for it in self.pushProtectionParameters:
                    writer.writeStartElement("Item")
                    writer.writeElementString("ProtectionType", int(it.protectionType))
                    writer.writeElementString(
                        "TransactionId", GXByteBuffer.hex(it.transactionId, False)
                    )
                    writer.writeElementString(
                        "OriginatorSystemTitle",
                        GXByteBuffer.hex(it.originatorSystemTitle, False),
                    )
                    writer.writeElementString(
                        "RecipientSystemTitle",
                        GXByteBuffer.hex(it.recipientSystemTitle, False),
                    )
                    writer.writeElementString(
                        "OtherInformation", GXByteBuffer.hex(it.otherInformation, False)
                    )
                    writer.writeElementString(
                        "DataProtectionKeyType", int(it.keyInfo.dataProtectionKeyType)
                    )
                    writer.writeStartElement("IdentifiedKey")
                    writer.writeElementString(
                        "KeyType", int(it.keyInfo.identifiedKey.keyType)
                    )
                    writer.writeEndElement()
                    writer.writeStartElement("WrappedKey")
                    writer.writeElementString(
                        "KeyType", int(it.keyInfo.wrappedKey.keyType)
                    )
                    writer.writeElementString(
                        "Key", GXByteBuffer.hex(it.keyInfo.wrappedKey.key, False)
                    )
                    writer.writeEndElement()
                    writer.writeStartElement("AgreedKey")
                    writer.writeElementString(
                        "Parameters",
                        GXByteBuffer.hex(it.keyInfo.agreedKey.parameters, False),
                    )
                    writer.writeElementString(
                        "Data", GXByteBuffer.hex(it.keyInfo.agreedKey.data, False)
                    )
                    writer.writeEndElement()
                    writer.writeEndElement()
                writer.writeEndElement()
            if self.version > 1:
                writer.writeElementString(
                    "PushOperationMethod", int(self.pushOperationMethod)
                )
                writer.writeElementString(
                    "ConfirmationParametersStartDate",
                    self.confirmationParameters.startDate,
                )
                writer.writeElementString(
                    "ConfirmationParametersInterval",
                    self.confirmationParameters.interval,
                )
                writer.writeElementString(
                    "LastConfirmationDateTime", self.lastConfirmationDateTime
                )

    def postLoad(self, reader):
        target = None
        # Update port reference.
        if self.portReference:
            target = reader.objects.findByLN(
                ObjectType.NONE, self.portReference.logicalName
            )
        if target and target != self.portReference:
            self.portReference = target
        # Upload object list after load.
        if self.pushObjectList:
            objects = []
            for obj, co in self.pushObjectList:
                target = reader.objects.findByLN(obj.objectType, obj.logicalName)
                if target and target != obj:
                    obj = target
                objects.append((obj, co))
            self.pushObjectList.clear()
            self.pushObjectList.extend(objects)



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from .enums.StateError import StateError
from .enums.ExceptionServiceError import ExceptionServiceError

class GXDLMSExceptionResponse(Exception):
    """
    DLMS specific exception class that has error description available from getDescription method.
    """

    #
    # Constructor for Confirmed ServiceError.
    #
    # @param service
    # @param type
    # @param value
    #
    def __init__(self, error=None, type_=None, value=0):
        Exception.__init__(self, self.__getStateError(error) + ". " + self.__getServiceError(type_))
        self.exceptionStateError = error
        self.exceptionServiceError = type_
        self.serviceErrorValue = value

    @classmethod
    def __getStateError(cls, stateError):
        str_ = ""
        if stateError == StateError.SERVICE_NOT_ALLOWED:
            str_ = "Service not allowed"
        elif stateError == StateError.SERVICE_UNKNOWN:
            str_ = "Service unknown"
        return str_

    @classmethod
    def __getServiceError(cls, error):
        str_ = ""
        if error == ExceptionServiceError.OPERATION_NOT_POSSIBLE:
            str_ = "Operation not possible"
        elif error == ExceptionServiceError.SERVICE_NOT_SUPPORTED:
            str_ = "Service not supported"
        elif error == ExceptionServiceError.OTHER_REASON:
            str_ = "Other reason"
        elif error == ExceptionServiceError.PDU_TOO_LONG:
            str_ = "PDU is too long"
        elif error == ExceptionServiceError.DECIPHERING_ERROR:
            str_ = "Ciphering failed"
        elif error == ExceptionServiceError.INVOCATION_COUNTER_ERROR:
            str_ = "Invocation counter is invalid."
        return str_



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from .GXDLMSObject import GXDLMSObject
from .IGXDLMSBase import IGXDLMSBase
from ..enums import ErrorCode
from ..internal._GXCommon import _GXCommon
from .enums import Weekdays
from ..enums import ObjectType, DataType
from .GXDLMSScheduleEntry import GXDLMSScheduleEntry
from ..GXDate import GXDate
from ..GXTime import GXTime
from ..GXByteBuffer import GXByteBuffer
from ..GXBitString import GXBitString
from ..internal._GXLocalizer import _GXLocalizer

# pylint: disable=too-many-instance-attributes
class GXDLMSSchedule(GXDLMSObject, IGXDLMSBase):
    """
    Online help:
    http://www.gurux.fi/Gurux.DLMS.Objects.GXDLMSSchedule
    """
    def __init__(self, ln=None, sn=0):
        """
        Constructor.

        ln : Logical Name of the object.
        sn : Short Name of the object.
        """
        #pylint: disable=super-with-arguments
        super(GXDLMSSchedule, self).__init__(ObjectType.SCHEDULE, ln, sn)
        # Specifies the scripts to be executed at given times.
        self.entries = []

    def getValues(self):
        return [self.logicalName,
                self.entries]

    #
    # Add entry to entries list.
    #
    # client: DLMS client.
    # entry: Schedule entry.
    # Returns Action bytes.
    def insert(self, client, entry):
        data = GXByteBuffer()
        self.addEntry(entry, data)
        return client.method(self, 2, data.array(), DataType.STRUCTURE)

    #
    # Remove entry from entries list.
    #
    # client: DLMS client.
    # entry: Schedule entry.
    # Returns Action bytes.
    def delete(self, client, entry):
        data = GXByteBuffer()
        data.setUInt8(DataType.STRUCTURE)
        #Add structure size.
        data.setUInt8(2)
        #firstIndex
        _GXCommon.setData(None, data, DataType.UINT16, entry.index)
        #lastIndex
        _GXCommon.setData(None, data, DataType.UINT16, entry.index)
        return client.method(self, 3, data.array(), DataType.STRUCTURE)

    #
    # Enable entry from entries list.
    #
    # client: DLMS client.
    # entry: Schedule entries.
    # Returns Action bytes.
    def enable(self, client, entry):
        data = GXByteBuffer()
        data.setUInt8(DataType.STRUCTURE)
        #Add structure size.
        data.setUInt8(4)
        #firstIndex
        _GXCommon.setData(None, data, DataType.UINT16, entry.index)
        #lastIndex
        _GXCommon.setData(None, data, DataType.UINT16, entry.index)
        _GXCommon.setData(None, data, DataType.UINT16, 0)
        _GXCommon.setData(None, data, DataType.UINT16, 0)
        return client.method(self, 1, data.array(), DataType.STRUCTURE)

    #
    # Disable entry from entries list.
    #
    # client: DLMS client.
    # entry: Schedule entries.
    # Returns Action bytes.
    def disable(self, client, entry):
        data = GXByteBuffer()
        data.setUInt8(DataType.STRUCTURE)
        #Add structure size.
        data.setUInt8(4)
        #firstIndex
        _GXCommon.setData(None, data, DataType.UINT16, 0)
        _GXCommon.setData(None, data, DataType.UINT16, 0)
        _GXCommon.setData(None, data, DataType.UINT16, entry.index)
        #lastIndex
        _GXCommon.setData(None, data, DataType.UINT16, entry.index)
        return client.method(self, 1, data.array(), DataType.STRUCTURE)

    #
    # Returns collection of attributes to read.  If attribute is static
    #      and
    # already read or device is returned HW error it is not returned.
    #
    def getAttributeIndexToRead(self, all_):
        attributes = []
        #  LN is static and read only once.
        if all_ or not self.logicalName:
            attributes.append(1)
        #  Entries
        if all_ or not self.isRead(2):
            attributes.append(2)
        return attributes

    #
    # Returns amount of attributes.
    #
    def getAttributeCount(self):
        return 2

    #
    # Returns amount of methods.
    #
    def getMethodCount(self):
        return 3

    def getNames(self):
        return (_GXLocalizer.gettext("Logical name"),\
            _GXLocalizer.gettext("Entries"))

    def getMethodNames(self):
        return (_GXLocalizer.gettext("Enable/disable"),\
            _GXLocalizer.gettext("Insert"),\
            _GXLocalizer.gettext("Delete"))

    def getDataType(self, index):
        if index == 1:
            return DataType.OCTET_STRING
        if index == 2:
            return DataType.ARRAY
        raise ValueError("getDataType failed. Invalid attribute index.")

    @classmethod
    def addEntry(cls, it, data):
        data.setUInt8(DataType.STRUCTURE)
        data.setUInt8(10)
        #Add index.
        data.setUInt8(DataType.UINT16)
        data.setUInt16(it.index)
        #Add enable.
        data.setUInt8(DataType.BOOLEAN)
        data.setUInt8(it.enable)
        #Add logical Name.
        data.setUInt8(DataType.OCTET_STRING)
        data.setUInt8(6)
        data.set(_GXCommon.logicalNameToBytes(it.logicalName))
        #Add script selector.
        data.setUInt8(DataType.UINT16)
        data.setUInt16(it.scriptSelector)
        #Add switch time.
        _GXCommon.setData(None, data, DataType.OCTET_STRING, GXTime(it.switchTime))
        #Add validity window.
        data.setUInt8(DataType.UINT16)
        data.setUInt16(it.validityWindow)
        #Add exec week days.
        _GXCommon.setData(None, data, DataType.BITSTRING, GXBitString.toBitString(it.execWeekdays, 7))
        #Add exec spec days.
        _GXCommon.setData(None, data, DataType.BITSTRING, it.execSpecDays)
        #Add begin date.
        _GXCommon.setData(None, data, DataType.OCTET_STRING, GXDate(it.beginDate))
        #Add end date.
        _GXCommon.setData(None, data, DataType.OCTET_STRING, GXDate(it.endDate))

    #
    # Returns value of given attribute.
    #
    def getValue(self, settings, e):
        if e.index == 1:
            return _GXCommon.logicalNameToBytes(self.logicalName)
        if e.index == 2:
            data = GXByteBuffer()
            data.setUInt8(DataType.ARRAY)
            _GXCommon.setObjectCount(len(self.entries), data)
            for it in self.entries:
                self.addEntry(it, data)
            return data.array()
        e.error = ErrorCode.READ_WRITE_DENIED
        return None

    #
    # Create a new entry.
    #
    @classmethod
    def createEntry(cls, settings, it):
        item = GXDLMSScheduleEntry()
        item.index = it[0]
        item.enable = it[1]
        item.logicalName = _GXCommon.toLogicalName(it[2])
        item.scriptSelector = it[3]
        item.switchTime = _GXCommon.changeType(settings, it[4], DataType.TIME)
        item.validityWindow = it[5]
        item.execWeekdays = Weekdays(it[6].toInteger())
        item.execSpecDays = str(it[7])
        item.beginDate = _GXCommon.changeType(settings, it[8], DataType.DATE)
        item.endDate = _GXCommon.changeType(settings, it[9], DataType.DATE)
        return item

    #
    # Set value of given attribute.
    #
    def setValue(self, settings, e):
        if e.index == 1:
            self.logicalName = _GXCommon.toLogicalName(e.value)
        elif e.index == 2:
            self.entries = []
            if e.value:
                for it in e.value:
                    self.entries.append(self.createEntry(settings, it))
        else:
            e.error = ErrorCode.READ_WRITE_DENIED

    def load(self, reader):
        self.entries = []
        if reader.isStartElement("Entries", True):
            while reader.isStartElement("Item", True):
                it = GXDLMSScheduleEntry()
                it.index = reader.readElementContentAsInt("Index")
                it.enable = reader.readElementContentAsInt("Enable") != 0
                it.logicalName = reader.readElementContentAsString("LogicalName")
                it.scriptSelector = reader.readElementContentAsInt("ScriptSelector")
                it.switchTime = reader.readElementContentAsTime("SwitchTime")
                it.validityWindow = reader.readElementContentAsInt("ValidityWindow")
                it.execWeekdays = Weekdays(reader.readElementContentAsInt("ExecWeekdays"))
                it.execSpecDays = reader.readElementContentAsString("ExecSpecDays")
                it.beginDate = reader.readElementContentAsDate("BeginDate")
                it.endDate = reader.readElementContentAsDate("EndDate")
                self.entries.append(it)
            reader.readEndElement("Entries")

    def save(self, writer):
        if self.entries:
            writer.writeStartElement("Entries")
            for it in self.entries:
                writer.writeStartElement("Item")
                writer.writeElementString("Index", it.index)
                writer.writeElementString("Enable", it.enable)
                writer.writeElementString("LogicalName", it.logicalName)
                writer.writeElementString("ScriptSelector", it.scriptSelector)
                writer.writeElementString("SwitchTime", it.switchTime)
                writer.writeElementString("ValidityWindow", it.validityWindow)
                writer.writeElementString("ExecWeekdays", int(it.execWeekdays))
                writer.writeElementString("ExecSpecDays", it.execSpecDays)
                writer.writeElementString("BeginDate", it.beginDate)
                writer.writeElementString("EndDate", it.endDate)
                writer.writeEndElement()
            writer.writeEndElement()



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from .GXDLMSObject import GXDLMSObject
from .IGXDLMSBase import IGXDLMSBase
from ..enums import ErrorCode
from ..internal._GXCommon import _GXCommon
from ..GXByteBuffer import GXByteBuffer
from ..enums import ObjectType, DataType
from .enums import BaudRate
from .GXDLMSModemInitialisation import GXDLMSModemInitialisation
from ..internal._GXLocalizer import _GXLocalizer

# pylint: disable=too-many-instance-attributes
class GXDLMSModemConfiguration(GXDLMSObject, IGXDLMSBase):
    """
    Online help:
    http://www.gurux.fi/Gurux.DLMS.Objects.GXDLMSModemConfiguration
    """

    #
    # Constructor.
    #
    # @param ln
    # Logical Name of the object.
    # @param sn
    # Short Name of the object.
    #
    def __init__(self, ln=None, sn=0):
        # pylint: disable=super-with-arguments
        super(GXDLMSModemConfiguration, self).__init__(
            ObjectType.MODEM_CONFIGURATION, ln, sn
        )
        self.initialisationStrings = []
        self.communicationSpeed = BaudRate.BAUDRATE_300
        self.modemProfile = self.defultProfiles()

    @classmethod
    def defultProfiles(cls):
        return [
            "OK",
            "CONNECT",
            "RING",
            "NO CARRIER",
            "ERROR",
            "CONNECT 1200",
            "NO DIAL TONE",
            "BUSY",
            "NO ANSWER",
            "CONNECT 600",
            "CONNECT 2400",
            "CONNECT 4800",
            "CONNECT 9600",
            "CONNECT 14 400",
            "CONNECT 28 800",
            "CONNECT 33 600",
            "CONNECT 56 000",
        ]

    def getValues(self):
        return [
            self.logicalName,
            self.communicationSpeed,
            self.initialisationStrings,
            self.modemProfile,
        ]

    #
    # Returns collection of attributes to read.  If attribute is static
    # and already read or device is returned HW error it is not returned.
    def getAttributeIndexToRead(self, all_):
        attributes = []
        #  LN is static and read only once.
        if all_ or not self.logicalName:
            attributes.append(1)
        #  CommunicationSpeed
        if all_ or not self.isRead(2):
            attributes.append(2)
        #  InitialisationStrings
        if all_ or not self.isRead(3):
            attributes.append(3)
        #  ModemProfile
        if all_ or not self.isRead(4):
            attributes.append(4)
        return attributes

    #
    # Returns amount of attributes.
    #
    def getAttributeCount(self):
        return 4

    #
    # Returns amount of methods.
    #
    def getMethodCount(self):
        return 0

    def getNames(self):
        return (_GXLocalizer.gettext("Logical name"),\
            _GXLocalizer.gettext("Communication speed"),\
            _GXLocalizer.gettext("Initialisation strings"),\
            _GXLocalizer.gettext("Modem profile"))

    def getDataType(self, index):
        if index == 1:
            return DataType.OCTET_STRING
        if index == 2:
            return DataType.ENUM
        if index == 3:
            return DataType.ARRAY
        if index == 4:
            return DataType.ARRAY
        raise ValueError("getDataType failed. Invalid attribute index.")

    #
    # Returns value of given attribute.
    #
    def getValue(self, settings, e):
        if e.index == 1:
            return _GXCommon.logicalNameToBytes(self.logicalName)
        if e.index == 2:
            return self.communicationSpeed
        if e.index == 3:
            data = GXByteBuffer()
            data.setUInt8(DataType.ARRAY)
            #  Add count
            cnt = 0
            if self.initialisationStrings:
                cnt = len(self.initialisationStrings)
            _GXCommon.setObjectCount(cnt, data)
            if cnt != 0:
                for it in self.initialisationStrings:
                    data.setUInt8(DataType.STRUCTURE)
                    data.setUInt8(3)
                    #  Count
                    _GXCommon.setData(
                        settings,
                        data,
                        DataType.OCTET_STRING,
                        _GXCommon.getBytes(it.request),
                    )
                    _GXCommon.setData(
                        settings,
                        data,
                        DataType.OCTET_STRING,
                        _GXCommon.getBytes(it.response),
                    )
                    _GXCommon.setData(settings, data, DataType.UINT16, it.delay)
            return data
        if e.index == 4:
            data = GXByteBuffer()
            data.setUInt8(DataType.ARRAY)
            #  Add count
            cnt = len(self.modemProfile)
            _GXCommon.setObjectCount(cnt, data)
            if cnt != 0:
                for it in self.modemProfile:
                    _GXCommon.setData(
                        settings, data, DataType.OCTET_STRING, _GXCommon.getBytes(it)
                    )
            return data
        e.error = ErrorCode.READ_WRITE_DENIED
        return None

    #
    # Set value of given attribute.
    #
    def setValue(self, settings, e):
        if e.index == 1:
            self.logicalName = _GXCommon.toLogicalName(e.value)
        elif e.index == 2:
            self.communicationSpeed = e.value
        elif e.index == 3:
            self.initialisationStrings = []
            if e.value:
                for it in e.value:
                    item = GXDLMSModemInitialisation()
                    item.request = it[0].decode("utf-8")
                    item.response = it[1].decode("utf-8")
                    if len(it) == 3:
                        item.delay = it[2]
                    self.initialisationStrings.append(item)
        elif e.index == 4:
            self.modemProfile = []
            if e.value:
                for it in e.value:
                    self.modemProfile.append(it.decode("utf-8"))
        else:
            e.error = ErrorCode.READ_WRITE_DENIED

    def load(self, reader):
        self.communicationSpeed = BaudRate(
            reader.readElementContentAsInt("CommunicationSpeed")
        )
        if reader.isStartElement("InitialisationStrings", True):
            while reader.isStartElement("Initialisation", True):
                it = GXDLMSModemInitialisation()
                it.request = reader.readElementContentAsString("Request")
                it.response = reader.readElementContentAsString("Response")
                it.delay = reader.readElementContentAsInt("Delay")
            reader.readEndElement("InitialisationStrings")
        self.modemProfile = reader.readElementContentAsString("ModemProfile", "").split(
            ";"
        )

    def save(self, writer):
        writer.writeElementString("CommunicationSpeed", int(self.communicationSpeed))
        writer.writeStartElement("InitialisationStrings")
        if self.initialisationStrings:
            for it in self.initialisationStrings:
                writer.writeStartElement("Initialisation")
                writer.writeElementString("Request", it.request)
                writer.writeElementString("Response", it.response)
                writer.writeElementString("Delay", it.delay)
                writer.writeEndElement()
        writer.writeEndElement()
        if self.modemProfile:
            writer.writeElementString("ModemProfile", ";".join(self.modemProfile))



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from __future__ import print_function
from .GXDLMSObject import GXDLMSObject
from .IGXDLMSBase import IGXDLMSBase
from ..internal._GXCommon import _GXCommon
from ..GXByteBuffer import GXByteBuffer
from ..enums import ErrorCode, ObjectType, DataType
from .enums import (
    SecuritySuite,
    CertificateEntity,
    CertificateType,
    SecurityPolicy0,
    SecurityPolicy,
)
from .GXDLMSCertificateInfo import GXDLMSCertificateInfo
from ..enums.Security import Security
from .enums.GlobalKeyType import GlobalKeyType
from ..internal._GXLocalizer import _GXLocalizer

# pylint: disable=too-many-public-methods,too-many-instance-attributes
class GXDLMSSecuritySetup(GXDLMSObject, IGXDLMSBase):
    """
    Online help:
    http://www.gurux.fi/Gurux.DLMS.Objects.GXDLMSSecuritySetup
    """

    def __init__(self, ln="0.0.43.0.0.255", sn=0):
        """
        Constructor.

        ln : Logical Name of the object.
        sn : Short Name of the object.
        """
        super(GXDLMSSecuritySetup, self).__init__(ObjectType.SECURITY_SETUP, ln, sn)
        self.version = 1
        self.securityPolicy = SecurityPolicy0.NOTHING
        # Security policy for version 1.
        self.securitySuite = SecuritySuite.SUITE_0
        self.certificates = []
        # Client system title.
        self.clientSystemTitle = None
        # Server system title.
        self.serverSystemTitle = None
        # Available certificates.
        self.certificates = []

    def getValues(self):
        return [
            self.logicalName,
            self.securityPolicy,
            self.securitySuite,
            self.clientSystemTitle,
            self.serverSystemTitle,
            self.certificates,
        ]

    #
    # Activates and strengthens the security policy.
    #
    # client: DLMS client that is used to generate action.
    # security: New security level.
    # Generated action.
    def activate(self, client, security):
        return client.method(self, 1, security, DataType.ENUM)

    #
    # Updates one or more global keys.
    #
    # client: DLMS client that is used to generate action.
    # kek: Master key, also known as Key Encrypting Key.
    # list: List of Global key types and keys.
    # Generated action.
    #
    def globalKeyTransfer(self, client, kek, list_):
        # pylint: disable=import-outside-toplevel
        from ..secure.GXDLMSSecureClient import GXDLMSSecureClient

        if not list_:
            raise ValueError("Invalid list. It is empty.")
        bb = GXByteBuffer()
        bb.setUInt8(DataType.ARRAY)
        bb.setUInt8(int(len(list_)))
        for it in list_:
            bb.setUInt8(DataType.STRUCTURE)
            bb.setUInt8(2)
            _GXCommon.setData(None, bb, DataType.ENUM, it[0])
            tmp = GXDLMSSecureClient.encrypt(kek, it[1])
            _GXCommon.setData(None, bb, DataType.OCTET_STRING, tmp)
        return client.method(self, 2, bb.array(), DataType.ARRAY)

    #
    # Agree on one or more symmetric keys using the key agreement
    #      algorithm.
    #
    # @param client
    # DLMS client that is used to generate action.
    # @param list
    # List of keys.
    # Generated action.
    #
    def __keyAgreement(self, client, list_):
        if not list_:
            raise ValueError("Invalid list. It is empty.")
        bb = GXByteBuffer()
        bb.setUInt8(DataType.ARRAY)
        bb.setUInt8(int(len(list_)))
        for it in list_:
            bb.setUInt8(DataType.STRUCTURE)
            bb.setUInt8(2)
            _GXCommon.setData(None, bb, DataType.ENUM, it.getKey())
            _GXCommon.setData(None, bb, DataType.OCTET_STRING, it)
        return client.method(self, 3, bb.array(), DataType.ARRAY)

    #
    # Agree on one or more symmetric keys using the key agreement
    #      algorithm.
    #
    # @param client
    # DLMS client that is used to generate action.
    # @param type
    # Global key type.
    # Generated action.
    def keyAgreement(self, client, type_):
        bb = GXByteBuffer()
        data = self.getEphemeralPublicKeyData(
            type_, client.ciphering.ephemeralKeyPair.public
        )
        bb.set(data, 1, 64)
        print("Signin public key: " + str(client.ciphering.signingKeyPair.public))
        # sign = GXASymmetric.getEphemeralPublicKeySignature(type_ ,
        # client.ciphering.ephemeralKeyPair.public,
        # client.ciphering.signingKeyPair.private)
        #  bb.set(sign)
        # print("Data: " + GXByteBuffer.hex(data))
        # print("Sign: " + GXByteBuffer.hex(sign))
        list_ = []
        list_.append((type_, bb.array()))
        return self.__keyAgreement(client, list_)

    def generateKeyPair(self, client, type_):
        return client.method(self, 4, type_, DataType.ENUM)

    def generateCertificate(self, client, type_):
        return client.method(self, 5, type_, DataType.ENUM)

    def importCertificate(self, client, certificate):
        # If certificate is a string.
        if isinstance(certificate, (str)):
            certificate = certificate.getEncoded()
        return client.method(self, 6, certificate, DataType.OCTET_STRING)

    def exportCertificateByEntity(self, client, entity, type_, systemTitle):
        if not systemTitle:
            raise ValueError("Invalid system title.")
        bb = GXByteBuffer()
        bb.setUInt8(DataType.STRUCTURE)
        bb.setUInt8(2)
        bb.setUInt8(DataType.ENUM)
        bb.setUInt8(0)
        bb.setUInt8(DataType.STRUCTURE)
        bb.setUInt8(3)
        bb.setUInt8(DataType.ENUM)
        bb.setUInt8(entity)
        bb.setUInt8(DataType.ENUM)
        bb.setUInt8(type_)
        _GXCommon.setData(None, bb, DataType.OCTET_STRING, systemTitle)
        return client.method(self, 7, bb.array(), DataType.STRUCTURE)

    def exportCertificateBySerial(self, client, serialNumber, issuer):
        bb = GXByteBuffer()
        bb.setUInt8(DataType.STRUCTURE)
        bb.setUInt8(2)
        bb.setUInt8(DataType.ENUM)
        bb.setUInt8(1)
        bb.setUInt8(DataType.STRUCTURE)
        bb.setUInt8(2)
        _GXCommon.setData(None, bb, DataType.OCTET_STRING, serialNumber.encode())
        _GXCommon.setData(None, bb, DataType.OCTET_STRING, issuer.encode())
        return client.method(self, 7, bb.array(), DataType.STRUCTURE)

    def removeCertificateByEntity(self, client, entity, type_, systemTitle):
        bb = GXByteBuffer()
        bb.setUInt8(DataType.STRUCTURE)
        bb.setUInt8(2)
        bb.setUInt8(DataType.ENUM)
        bb.setUInt8(0)
        bb.setUInt8(DataType.STRUCTURE)
        bb.setUInt8(3)
        bb.setUInt8(DataType.ENUM)
        bb.setUInt8(entity)
        bb.setUInt8(DataType.ENUM)
        bb.setUInt8(type_)
        _GXCommon.setData(None, bb, DataType.OCTET_STRING, systemTitle)
        return client.method(self, 8, bb.array(), DataType.STRUCTURE)

    def removeCertificateBySerial(self, client, serialNumber, issuer):
        bb = GXByteBuffer()
        bb.setUInt8(DataType.STRUCTURE)
        bb.setUInt8(2)
        bb.setUInt8(DataType.ENUM)
        bb.setUInt8(1)
        bb.setUInt8(DataType.STRUCTURE)
        bb.setUInt8(2)
        _GXCommon.setData(None, bb, DataType.OCTET_STRING, serialNumber.encode())
        _GXCommon.setData(None, bb, DataType.OCTET_STRING, issuer.encode())
        return client.method(self, 8, bb.array(), DataType.STRUCTURE)

    @classmethod
    def systemTitleToSubject(cls, systemTitle):
        bb = GXByteBuffer(systemTitle)
        bb.setUInt8(0, 0)
        bb.setUInt8(0, 1)
        bb.setUInt8(0, 2)
        subject = "CN=" + str(systemTitle, 0, 3)
        subject += str(bb.getUInt64())
        return subject

    def invoke(self, settings, e):
        # pylint: disable=bad-option-value,redefined-variable-type
        if e.index == 1:
            if self.version == 0:
                self.securityPolicy = SecurityPolicy0(e.parameters)
                if self.securityPolicy == SecurityPolicy0.AUTHENTICATED:
                    settings.cipher.security = Security.AUTHENTICATION
                elif self.securityPolicy == SecurityPolicy0.ENCRYPTED:
                    settings.cipher.security = Security.ENCRYPTION
                elif self.securityPolicy == SecurityPolicy0.AUTHENTICATED_ENCRYPTED:
                    settings.cipher.security = Security.AUTHENTICATION_ENCRYPTION
            elif self.version == 1:
                self.securityPolicy = SecurityPolicy(e.parameters)
                if self.securityPolicy & SecurityPolicy.AUTHENTICATED_RESPONSE != 0:
                    settings.cipher.security = Security(
                        settings.cipher.security | Security.AUTHENTICATION
                    )
                if self.securityPolicy & SecurityPolicy.ENCRYPTED_RESPONSE != 0:
                    settings.cipher.security = Security(
                        settings.cipher.security | Security.ENCRYPTION
                    )
        elif e.index == 2:
            # pylint: disable=import-outside-toplevel
            from ..secure.GXDLMSSecureClient import GXDLMSSecureClient

            # if settings.Cipher is null non secure server is used.
            # Keys are take in action after reply is generated.
            for tmp in e.parameters:
                item = tmp
                type_ = GlobalKeyType(item[0])
                if type_ == GlobalKeyType.UNICAST_ENCRYPTION:
                    GXDLMSSecureClient.decrypt(settings.kek, item[1])
                elif type_ == GlobalKeyType.BROADCAST_ENCRYPTION:
                    e.error = ErrorCode.READ_WRITE_DENIED
                elif type_ == GlobalKeyType.AUTHENTICATION:
                    GXDLMSSecureClient.decrypt(settings.kek, item[1])
                elif type_ == GlobalKeyType.KEK:
                    GXDLMSSecureClient.decrypt(settings.kek, item[1])
                else:
                    e.error = ErrorCode.READ_WRITE_DENIED
        elif e.index == 3:
            e.rror = ErrorCode.HARDWARE_FAULT
        elif e.index == 4:
            e.rror = ErrorCode.HARDWARE_FAULT
        elif e.index == 5:
            e.rror = ErrorCode.HARDWARE_FAULT
        elif e.index == 6:
            e.rror = ErrorCode.HARDWARE_FAULT
        elif e.index == 8:
            e.error = ErrorCode.READ_WRITE_DENIED
        else:
            e.error = ErrorCode.READ_WRITE_DENIED

    @classmethod
    def applyKeys(cls, settings, e):
        # pylint: disable=import-outside-toplevel
        from ..secure.GXDLMSSecureClient import GXDLMSSecureClient

        for tmp in e.parameters:
            item = tmp
            type_ = GlobalKeyType(item[0])
            data = GXDLMSSecureClient.decrypt(settings.kek, item[1])
            if type_ == GlobalKeyType.UNICAST_ENCRYPTION:
                settings.cipher.blockCipherKey = data
            elif type_ == GlobalKeyType.BROADCAST_ENCRYPTION:
                e.error = ErrorCode.READ_WRITE_DENIED
            elif type_ == GlobalKeyType.AUTHENTICATION:
                settings.cipher.authenticationKey = data
            elif type_ == GlobalKeyType.KEK:
                settings.kek = data
            else:
                e.error = ErrorCode.READ_WRITE_DENIED

    def getAttributeIndexToRead(self, all_):
        attributes = []
        if all_ or not self.logicalName:
            attributes.append(1)
        if all_ or self.canRead(2):
            attributes.append(2)
        if all_ or self.canRead(3):
            attributes.append(3)
        if all_ or self.canRead(4):
            attributes.append(4)
        if all_ or self.canRead(5):
            attributes.append(5)
        if self.version != 0:
            if all_ or self.canRead(6):
                attributes.append(6)
        return attributes

    def getAttributeCount(self):
        if self.version == 0:
            return 5
        return 6

    def getMethodCount(self):
        if self.version == 0:
            return 2
        return 8

    def getNames(self):
        return (_GXLocalizer.gettext("Logical name"),\
            _GXLocalizer.gettext("Security policy"),\
            _GXLocalizer.gettext("Security suite"),\
            _GXLocalizer.gettext("Client system title"),\
            _GXLocalizer.gettext("Server system title"),\
            _GXLocalizer.gettext("Certificates"))

    def getMethodNames(self):
        return (_GXLocalizer.gettext("Adjust to quarter"),\
            _GXLocalizer.gettext("Adjust to measuring period"),\
            _GXLocalizer.gettext("Adjust to minute"),\
            _GXLocalizer.gettext("Adjust to preset time"),\
            _GXLocalizer.gettext("Preset adjusting time"),\
            _GXLocalizer.gettext("Shift time"))

    def getDataType(self, index):
        if index == 1:
            ret = DataType.OCTET_STRING
        elif index == 2:
            ret = DataType.ENUM
        elif index == 3:
            ret = DataType.ENUM
        elif index == 4:
            ret = DataType.OCTET_STRING
        elif index == 5:
            ret = DataType.OCTET_STRING
        elif self.version > 0:
            if index == 6:
                ret = DataType.ARRAY
            else:
                raise ValueError("getDataType failed. Invalid attribute index.")
        else:
            raise ValueError("getDataType failed. Invalid attribute index.")
        return ret

    @classmethod
    def getCertificatesByteArray(cls, settings):
        bb = GXByteBuffer()
        bb.setUInt8(DataType.ARRAY)
        if settings.cipher.certificates:
            _GXCommon.setObjectCount(len(settings.cipher.certificates), bb)
            for it in settings.cipher.certificates:
                bb.setUInt8(DataType.STRUCTURE)
                _GXCommon.setObjectCount(6, bb)
                bb.setUInt8(DataType.ENUM)
                bb.setUInt8(CertificateEntity.SERVER)
                bb.setUInt8(DataType.ENUM)
                bb.setUInt8(CertificateType.DIGITAL_SIGNATURE)
                _GXCommon.addString(it.serialNumber, bb)
                _GXCommon.addString(it.issuer, bb)
                _GXCommon.addString(it.subject, bb)
                _GXCommon.addString("", bb)
        else:
            bb.setUInt8(0)
        return bb.array()

    def getValue(self, settings, e):
        # pylint: disable=bad-option-value,redefined-variable-type
        if e.index == 1:
            ret = _GXCommon.logicalNameToBytes(self.logicalName)
        elif e.index == 2:
            ret = self.securityPolicy
        elif e.index == 3:
            ret = self.securitySuite
        elif e.index == 4:
            ret = self.clientSystemTitle
        elif e.index == 5:
            ret = self.serverSystemTitle
        elif e.index == 6:
            ret = self.getCertificatesByteArray(settings)
        else:
            e.error = ErrorCode.READ_WRITE_DENIED
        return ret

    def updateSertificates(self, list_):
        self.certificates = []
        if list_:
            for it in list_:
                info = GXDLMSCertificateInfo()
                info.entity = CertificateEntity(it[0])
                info.type_ = CertificateType(it[1])
                info.serialNumber = it[2]
                info.issuer = it[3]
                info.subject = it[4]
                info.subjectAltName = it[5]
                self.certificates.append(info)

    def setValue(self, settings, e):
        # pylint: disable=bad-option-value,redefined-variable-type
        if e.index == 1:
            self.logicalName = _GXCommon.toLogicalName(e.value)
        elif e.index == 2:
            if settings.isServer:
                e.error = ErrorCode.READ_WRITE_DENIED
            else:
                if self.version == 0:
                    self.securityPolicy = SecurityPolicy0(e.value)
                else:
                    self.securityPolicy = SecurityPolicy(e.value)
        elif e.index == 3:
            self.securitySuite = e.value
        elif e.index == 4:
            self.clientSystemTitle = e.value
        elif e.index == 5:
            self.serverSystemTitle = e.value
        elif e.index == 6:
            self.updateSertificates(e.value)
        else:
            e.error = ErrorCode.READ_WRITE_DENIED

    @classmethod
    def getEphemeralPublicKeyData(cls, keyId, ephemeralKey):
        # pylint: disable=unused-argument
        # tmp =
        # (GXAsn1Converter.fromByteArray(ephemeralKey.getEncoded())).get(1)
        # epk = GXByteBuffer(tmp.value)
        # epk.setUInt8(int(keyId), 0)
        # return epk
        return None

    def load(self, reader):
        self.securityPolicy = SecurityPolicy(
            reader.readElementContentAsInt("SecurityPolicy")
        )
        self.securitySuite = SecuritySuite(
            reader.readElementContentAsInt("SecuritySuite")
        )
        str_ = reader.readElementContentAsString("ClientSystemTitle")
        if str_ is None:
            self.clientSystemTitle = None
        else:
            self.clientSystemTitle = GXByteBuffer.hexToBytes(str_)
        str_ = reader.readElementContentAsString("ServerSystemTitle")
        if str_ is None:
            self.serverSystemTitle = None
        else:
            self.serverSystemTitle = GXByteBuffer.hexToBytes(str_)
        self.certificates = []
        if reader.isStartElement("Certificates", True):
            while reader.isStartElement("Item", True):
                it = GXDLMSCertificateInfo()
                self.certificates.append(it)
                it.entity = CertificateEntity(reader.readElementContentAsInt("Entity"))
                it.type_ = CertificateType(reader.readElementContentAsInt("Type"))
                it.serialNumber = reader.readElementContentAsString("SerialNumber")
                it.issuer = reader.readElementContentAsString("Issuer")
                it.subject = reader.readElementContentAsString("Subject")
                it.subjectAltName = reader.readElementContentAsString("SubjectAltName")
            reader.readEndElement("Certificates")

    def save(self, writer):
        writer.writeElementString("SecurityPolicy", int(self.securityPolicy))
        writer.writeElementString("SecuritySuite", int(self.securitySuite))
        writer.writeElementString(
            "ClientSystemTitle", GXByteBuffer.hex(self.clientSystemTitle)
        )
        writer.writeElementString(
            "ServerSystemTitle", GXByteBuffer.hex(self.serverSystemTitle)
        )
        writer.writeStartElement("Certificates")
        if self.certificates:
            for it in self.certificates:
                writer.writeStartElement("Item")
                writer.writeElementString("Entity", it.entity)
                writer.writeElementString("Type", it.type_)
                writer.writeElementString("SerialNumber", it.serialNumber)
                writer.writeElementString("Issuer", it.issuer)
                writer.writeElementString("Subject", it.subject)
                writer.writeElementString("SubjectAltName", it.subjectAltName)
                writer.writeEndElement()
        writer.writeEndElement()



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename:        $HeadURL$
#
#  Version:         $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from abc import ABCMeta, abstractmethod

ABC = ABCMeta('ABC', (object,), {'__slots__': ()})

# pylint: disable=too-many-public-methods
class GXICipher(ABC):
    #
    # Reset encrypt settings.
    #
    @abstractmethod
    def reset(self):
        raise ValueError("abstract method is called.")

    # Is ciphering used.
    @abstractmethod
    def isCiphered(self):
        raise ValueError("abstract method is called.")

    #
    # Used security.
    #
    @abstractmethod
    def getSecurity(self):
        raise ValueError("abstract method is called.")

    #
    # @param value
    #            Used security.
    #
    @abstractmethod
    def setSecurity(self, value):
        raise ValueError("abstract method is called.")

    #
    # System title.
    #
    @abstractmethod
    def getSystemTitle(self):
        raise ValueError("abstract method is called.")

    #
    # Recipient system Title.
    #
    @abstractmethod
    def getRecipientSystemTitle(self):
        raise ValueError("abstract method is called.")

    #
    # Block cipher key.
    #
    @abstractmethod
    def getBlockCipherKey(self):
        raise ValueError("abstract method is called.")

    #
    # Authentication key.
    #
    @abstractmethod
    def getAuthenticationKey(self):
        raise ValueError("abstract method is called.")

    #
    # @param value
    #            Authentication key.
    #
    @abstractmethod
    def setAuthenticationKey(self, value):
        raise ValueError("abstract method is called.")

    #
    # Invocation counter.
    #
    @abstractmethod
    def getInvocationCounter(self):
        raise ValueError("abstract method is called.")

    #
    # Used security suite.
    #
    @abstractmethod
    def getSecuritySuite(self):
        raise ValueError("abstract method is called.")

    #
    # Ephemeral key pair.
    #
    @abstractmethod
    def getEphemeralKeyPair(self):
        raise ValueError("abstract method is called.")

    #
    # @param value
    #            Ephemeral key pair.
    #
    @abstractmethod
    def setEphemeralKeyPair(self, value):
        raise ValueError("abstract method is called.")

    #
    # Client's key agreement key pair.
    #
    @abstractmethod
    def getKeyAgreementKeyPair(self):
        raise ValueError("abstract method is called.")

    #
    # @param value
    #            Client's key agreement key pair.
    #
    @abstractmethod
    def setKeyAgreementKeyPair(self, value):
        raise ValueError("abstract method is called.")

    #
    # Target (Server or client) Public key.
    #
    @abstractmethod
    def getPublicKeys(self):
        raise ValueError("abstract method is called.")

    #
    # Available certificates.
    #
    @abstractmethod
    def getCertificates(self):
        raise ValueError("abstract method is called.")

    #
    # Signing key pair.
    #
    @abstractmethod
    def getSigningKeyPair(self):
        raise ValueError("abstract method is called.")

    #
    # @param value
    #            Signing key pair.
    #
    @abstractmethod
    def setSigningKeyPair(self, value):
        raise ValueError("abstract method is called.")

    #
    # Shared secret is generated when connection is made.
    #
    @abstractmethod
    def getSharedSecret(self):
        raise ValueError("abstract method is called.")

    #
    # @param value
    #            Shared secret is generated when connection is made.
    #
    @abstractmethod
    def setSharedSecret(self, value):
        raise ValueError("abstract method is called.")

    #
    # Dedicated key.
    #
    @abstractmethod
    def getDedicatedKey(self):
        raise ValueError("abstract method is called.")

    #
    # @param value
    #            Dedicated key.
    #
    @abstractmethod
    def setDedicatedKey(self, value):
        raise ValueError("abstract method is called.")



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http:#www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http:#www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from gurux_dlms.objects.enums.GainResolution import GainResolution
from gurux_dlms.objects.enums.Modulation import Modulation


class GXDLMSNeighbourTable:
    """
    The neighbour table contains information about all the devices within the POS of the device
    """

    __shortAddress = None
    "MAC Short Address of the node."

    __enabled = None
    "Is Payload Modulation scheme used."

    __toneMap = ""
    "Frequency sub-band can be used for communication with the device."

    __modulation = Modulation.ROBUST_MODE
    "Modulation type."

    __txGain = None
    "Tx Gain."

    __txRes = GainResolution.D_B6
    "Tx Gain resolution."

    __txCoeff = ""
    "Transmitter gain for each group of tones represented by one valid bit of the tone map."

    __lqi = 0
    "Link Quality Indicator."

    __phaseDifferential = None
    "Phase difference in multiples of 60 degrees between the mains phase of the local node and the neighbour node."

    __tMRValidTime = 0
    "Remaining time in minutes until which the tone map response parameters in the neighbour table are considered valid."

    __neighbourValidTime = 0
    "Remaining time in minutes until which this entry in the neighbour table is considered valid."

    @property
    def shortAddress(self):
        """
        MAC Short Address of the node.
        """
        return self.__shortAddress

    @shortAddress.setter
    def shortAddress(self, value):
        self.__shortAddress = value

    @property
    def enabled(self):
        """
        Is Payload Modulation scheme used.
        """
        return self.__enabled

    @enabled.setter
    def enabled(self, value):
        self.__enabled = value

    @property
    def toneMap(self):
        """
        Frequency sub-band can be used for communication with the device.
        """
        return self.__toneMap

    @toneMap.setter
    def toneMap(self, value):
        self.__toneMap = value

    @property
    def modulation(self):
        """
        Modulation type.
        """
        return self.__modulation

    @modulation.setter
    def modulation(self, value):
        self.__modulation = value

    @property
    def txGain(self):
        """
        Tx Gain.
        """
        return self.__txGain

    @txGain.setter
    def txGain(self, value):
        self.__txGain = value

    @property
    def txRes(self):
        """
        Tx Gain resolution.
        """
        return self.__txRes

    @txRes.setter
    def txRes(self, value):
        self.__txRes = value

    @property
    def txCoeff(self):
        """
        Transmitter gain for each group of tones represented by one valid bit of the tone map.
        """
        return self.__txCoeff

    @txCoeff.setter
    def txCoeff(self, value):
        self.__txCoeff = value

    @property
    def lqi(self):
        """
        Link Quality Indicator.
        """
        return self.__lqi

    @lqi.setter
    def lqi(self, value):
        self.__lqi = value

    @property
    def phaseDifferential(self):
        """
        Phase difference in multiples of 60 degrees between the mains phase of the local node and the neighbour node.
        """
        return self.__phaseDifferential

    @phaseDifferential.setter
    def phaseDifferential(self, value):
        self.__phaseDifferential = value

    @property
    def tMRValidTime(self):
        """
        Remaining time in minutes until which the tone map response parameters in the neighbour table are considered valid.
        """
        return self.__tMRValidTime

    @tMRValidTime.setter
    def tMRValidTime(self, value):
        self.__tMRValidTime = value

    @property
    def neighbourValidTime(self):
        """
        Remaining time in minutes until which this entry in the neighbour table is considered valid.
        """
        return self.__neighbourValidTime

    @neighbourValidTime.setter
    def neighbourValidTime(self, value):
        self.__neighbourValidTime = value
